/* Generated by Cython 0.19.2 on Tue Oct 15 11:13:27 2013 */

#define PY_SSIZE_T_CLEAN
#ifndef CYTHON_USE_PYLONG_INTERNALS
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 0
#else
#include "pyconfig.h"
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 1
#else
#define CYTHON_USE_PYLONG_INTERNALS 0
#endif
#endif
#endif
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define __Pyx_PyIndex_Check(o) (PyNumber_Check(o) && !PyFloat_Check(o) && \
                                  !PyComplex_Check(o))
  #define PyIndex_Check __Pyx_PyIndex_Check
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
  #define __Pyx_PyIndex_Check PyIndex_Check
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_TPFLAGS_HAVE_VERSION_TAG 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj) || \
                                         PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (Py_TYPE(obj) == &PyBaseString_Type)
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  /* Initialize NaN. The sign is irrelevant, an exponent with all bits 1 and
   a nonzero mantissa means NaN. If the first bit in the mantissa is 1, it is
   a quiet NaN. */
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE___gappedbuf
#define __PYX_HAVE_API___gappedbuf
#include "string.h"
#include "stdlib.h"
#include "stdio.h"
#include "pythread.h"
#include "_gappedbufre.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromUString(s)  __Pyx_PyObject_FromString((char*)s)
#define __Pyx_PyBytes_FromUString(s)   __Pyx_PyBytes_FromString((char*)s)
#define __Pyx_PyStr_FromUString(s)     __Pyx_PyStr_FromString((char*)s)
#define __Pyx_PyUnicode_FromUString(s) __Pyx_PyUnicode_FromString((char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return u_end - u - 1;
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params() {
    PyObject* sys = NULL;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    sys = PyImport_ImportModule("sys");
    if (sys == NULL) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    if (default_encoding == NULL) goto bad;
    if (strcmp(PyBytes_AsString(default_encoding), "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        const char* default_encoding_c = PyBytes_AS_STRING(default_encoding);
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (ascii_chars_u == NULL) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (ascii_chars_b == NULL || strncmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
    }
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return 0;
bad:
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params() {
    PyObject* sys = NULL;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (sys == NULL) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    if (default_encoding == NULL) goto bad;
    default_encoding_c = PyBytes_AS_STRING(default_encoding);
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(sys);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(sys);
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "_gappedbuf.pyx",
  "_gappedbuf.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct GappedBufferObj;

/* "_gappedbuf.pxd":2
 * 
 * cdef public class GappedBuffer [object GappedBufferObj, type GappedBufferType]:             # <<<<<<<<<<<<<<
 *     cdef public Py_ssize_t bufsize
 *     cdef public Py_ssize_t numelems
 */
struct GappedBufferObj {
  PyObject_HEAD
  struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *__pyx_vtab;
  Py_ssize_t bufsize;
  Py_ssize_t numelems;
  Py_ssize_t gap;
  Py_ssize_t gapsize;
  Py_UCS4 *buf;
};

__PYX_EXTERN_C DL_EXPORT(PyTypeObject) GappedBufferType;

__PYX_EXTERN_C DL_EXPORT(PyTypeObject) GappedBufferType;

/* "_gappedbuf.pyx":18
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()
 * 
 * cdef class GappedBuffer:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         DEF BUFSIZE = 4096
 */

struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer {
  void (*_resize_buf)(struct GappedBufferObj *, Py_ssize_t);
  void (*_expand_gap)(struct GappedBufferObj *, Py_ssize_t);
  void (*_shrink_gap)(struct GappedBufferObj *);
  void (*_move_gap)(struct GappedBufferObj *, Py_ssize_t);
  Py_UCS4 (*_get_max)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t);
  void (*_copy_chars1)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t);
  void (*_copy_chars2)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t);
  void (*_copy_chars4)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t);
  int (*_get_slice)(struct GappedBufferObj *, PyObject *, Py_ssize_t *, Py_ssize_t *);
  PyObject *(*_insert)(struct GappedBufferObj *, Py_ssize_t, PyObject *);
  PyObject *(*insert)(struct GappedBufferObj *, Py_ssize_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct GappedBufferObj *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_delete)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t);
  PyObject *(*delete)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch);
  PyObject *(*replace)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch);
  PyObject *(*getint)(struct GappedBufferObj *, Py_ssize_t, int __pyx_skip_dispatch);
  PyObject *(*getints)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch);
  PyObject *(*_insertints)(struct GappedBufferObj *, Py_ssize_t, PyObject *);
  PyObject *(*insertints)(struct GappedBufferObj *, Py_ssize_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*appendints)(struct GappedBufferObj *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*replaceints)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*setints)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, unsigned long, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *__pyx_vtabptr_10_gappedbuf_GappedBuffer;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_XDECREF_SET(r, v) do {                            \
        PyObject *tmp = (PyObject *) r;                         \
        r = v; __Pyx_XDECREF(tmp);                              \
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {                             \
        PyObject *tmp = (PyObject *) r;                         \
        r = v; __Pyx_DECREF(tmp);                               \
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed); /*proto*/

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename); /*proto*/

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact); /*proto*/

#define __Pyx_GetItemInt(o, i, size, to_py_func, is_list, wraparound, boundscheck) \
    (((size) <= sizeof(Py_ssize_t)) ? \
    __Pyx_GetItemInt_Fast(o, i, is_list, wraparound, boundscheck) : \
    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
#define __Pyx_GetItemInt_List(o, i, size, to_py_func, is_list, wraparound, boundscheck) \
    (((size) <= sizeof(Py_ssize_t)) ? \
    __Pyx_GetItemInt_List_Fast(o, i, wraparound, boundscheck) : \
    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, size, to_py_func, is_list, wraparound, boundscheck) \
    (((size) <= sizeof(Py_ssize_t)) ? \
    __Pyx_GetItemInt_Tuple_Fast(o, i, wraparound, boundscheck) : \
    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

static int __Pyx_SetVtable(PyObject *dict, void *vtable); /*proto*/

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level); /*proto*/

static CYTHON_INLINE Py_UCS4 __Pyx_PyObject_AsPy_UCS4(PyObject*);

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'rtdef' */

/* Module declarations from '_gappedbuf' */
static PyTypeObject *__pyx_ptype_10_gappedbuf_GappedBuffer = 0;
#define __Pyx_MODULE_NAME "_gappedbuf"
int __pyx_module_is_main__gappedbuf = 0;

/* Implementation of '_gappedbuf' */
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_pf_10_gappedbuf_init_gappedbufre(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer___cinit__(struct GappedBufferObj *__pyx_v_self); /* proto */
static void __pyx_pf_10_gappedbuf_12GappedBuffer_2__dealloc__(struct GappedBufferObj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_4insert(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_6append(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_8delete(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_10replace(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_12get(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end); /* proto */
static Py_ssize_t __pyx_pf_10_gappedbuf_12GappedBuffer_14__len__(struct GappedBufferObj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_16__getitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_18__setitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_20__delitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_22findchr(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_c, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_24rfindchr(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_c, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_26getint(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_pos); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_28getints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_30insertints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_32appendints(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_34replaceints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_36setints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, unsigned long __pyx_v_v); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_38findint(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_nums, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_v_comp_ne); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_40rfindint(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_nums, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_v_comp_ne); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize___get__(struct GappedBufferObj *__pyx_v_self); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_8numelems___get__(struct GappedBufferObj *__pyx_v_self); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_8numelems_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_3gap___get__(struct GappedBufferObj *__pyx_v_self); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_3gap_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize___get__(struct GappedBufferObj *__pyx_v_self); /* proto */
static int __pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_tp_new_10_gappedbuf_GappedBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static char __pyx_k_1[] = "Invalid index value: %d, %d";
static char __pyx_k_2[] = "Invalid range: %d, %d, %d";
static char __pyx_k_3[] = "step is not supported";
static char __pyx_k_5[] = "Invalid string length: %d";
static char __pyx_k_8[] = "/Users/ishimoto/src/kaa/_gappedbuf/_gappedbuf.pyx";
static char __pyx_k__c[] = "c";
static char __pyx_k__s[] = "s";
static char __pyx_k__v[] = "v";
static char __pyx_k__end[] = "end";
static char __pyx_k__get[] = "get";
static char __pyx_k__sys[] = "sys";
static char __pyx_k__nums[] = "nums";
static char __pyx_k__begin[] = "begin";
static char __pyx_k__index[] = "index";
static char __pyx_k__append[] = "append";
static char __pyx_k__delete[] = "delete";
static char __pyx_k__getint[] = "getint";
static char __pyx_k__insert[] = "insert";
static char __pyx_k__comp_ne[] = "comp_ne";
static char __pyx_k__getints[] = "getints";
static char __pyx_k__modules[] = "modules";
static char __pyx_k__replace[] = "replace";
static char __pyx_k__setints[] = "setints";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__IndexError[] = "IndexError";
static char __pyx_k__ValueError[] = "ValueError";
static char __pyx_k____import__[] = "__import__";
static char __pyx_k___gappedbuf[] = "_gappedbuf";
static char __pyx_k__appendints[] = "appendints";
static char __pyx_k__insertints[] = "insertints";
static char __pyx_k__MemoryError[] = "MemoryError";
static char __pyx_k__replaceints[] = "replaceints";
static char __pyx_k__RuntimeError[] = "RuntimeError";
static char __pyx_k___gappedbufre[] = "_gappedbufre";
static char __pyx_k____pyx_vtable__[] = "__pyx_vtable__";
static char __pyx_k__init_gappedbufre[] = "init_gappedbufre";
static PyObject *__pyx_kp_s_1;
static PyObject *__pyx_kp_s_2;
static PyObject *__pyx_kp_s_3;
static PyObject *__pyx_kp_s_5;
static PyObject *__pyx_kp_s_8;
static PyObject *__pyx_n_s__IndexError;
static PyObject *__pyx_n_s__MemoryError;
static PyObject *__pyx_n_s__RuntimeError;
static PyObject *__pyx_n_s__ValueError;
static PyObject *__pyx_n_s____import__;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____pyx_vtable__;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s___gappedbuf;
static PyObject *__pyx_n_s___gappedbufre;
static PyObject *__pyx_n_s__append;
static PyObject *__pyx_n_s__appendints;
static PyObject *__pyx_n_s__begin;
static PyObject *__pyx_n_s__c;
static PyObject *__pyx_n_s__comp_ne;
static PyObject *__pyx_n_s__delete;
static PyObject *__pyx_n_s__end;
static PyObject *__pyx_n_s__get;
static PyObject *__pyx_n_s__getint;
static PyObject *__pyx_n_s__getints;
static PyObject *__pyx_n_s__index;
static PyObject *__pyx_n_s__init_gappedbufre;
static PyObject *__pyx_n_s__insert;
static PyObject *__pyx_n_s__insertints;
static PyObject *__pyx_n_s__modules;
static PyObject *__pyx_n_s__nums;
static PyObject *__pyx_n_s__replace;
static PyObject *__pyx_n_s__replaceints;
static PyObject *__pyx_n_s__s;
static PyObject *__pyx_n_s__setints;
static PyObject *__pyx_n_s__sys;
static PyObject *__pyx_n_s__v;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k_tuple_4;
static PyObject *__pyx_k_tuple_6;
static PyObject *__pyx_k_codeobj_7;

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_1init_gappedbufre(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10_gappedbuf_1init_gappedbufre = {__Pyx_NAMESTR("init_gappedbufre"), (PyCFunction)__pyx_pw_10_gappedbuf_1init_gappedbufre, METH_NOARGS, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_10_gappedbuf_1init_gappedbufre(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_gappedbufre (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_init_gappedbufre(__pyx_self);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":14
 * 
 * 
 * def init_gappedbufre():             # <<<<<<<<<<<<<<
 *     import sys
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()
 */

static PyObject *__pyx_pf_10_gappedbuf_init_gappedbufre(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_sys = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_gappedbufre", 0);

  /* "_gappedbuf.pyx":15
 * 
 * def init_gappedbufre():
 *     import sys             # <<<<<<<<<<<<<<
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()
 * 
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__sys), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sys = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_gappedbuf.pyx":16
 * def init_gappedbufre():
 *     import sys
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()             # <<<<<<<<<<<<<<
 * 
 * cdef class GappedBuffer:
 */
  __pyx_t_1 = PyInit__gappedbufre(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 16; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s__modules); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 16; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s___gappedbufre), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 16; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_gappedbuf.init_gappedbufre", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer___cinit__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":19
 * 
 * cdef class GappedBuffer:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         DEF BUFSIZE = 4096
 * 
 */

static int __pyx_pf_10_gappedbuf_12GappedBuffer___cinit__(struct GappedBufferObj *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "_gappedbuf.pyx":22
 *         DEF BUFSIZE = 4096
 * 
 *         self.buf = <Py_UCS4*>malloc(BUFSIZE * sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *         if not self.buf:
 *             raise MemoryError()
 */
  __pyx_v_self->buf = ((Py_UCS4 *)malloc((4096 * (sizeof(Py_UCS4)))));

  /* "_gappedbuf.pyx":23
 * 
 *         self.buf = <Py_UCS4*>malloc(BUFSIZE * sizeof(Py_UCS4))
 *         if not self.buf:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->buf != 0)) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":24
 *         self.buf = <Py_UCS4*>malloc(BUFSIZE * sizeof(Py_UCS4))
 *         if not self.buf:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self.bufsize = BUFSIZE
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 24; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":26
 *             raise MemoryError()
 * 
 *         self.bufsize = BUFSIZE             # <<<<<<<<<<<<<<
 *         self.numelems = 0
 *         self.gap = 0
 */
  __pyx_v_self->bufsize = 4096;

  /* "_gappedbuf.pyx":27
 * 
 *         self.bufsize = BUFSIZE
 *         self.numelems = 0             # <<<<<<<<<<<<<<
 *         self.gap = 0
 *         self.gapsize = BUFSIZE
 */
  __pyx_v_self->numelems = 0;

  /* "_gappedbuf.pyx":28
 *         self.bufsize = BUFSIZE
 *         self.numelems = 0
 *         self.gap = 0             # <<<<<<<<<<<<<<
 *         self.gapsize = BUFSIZE
 * 
 */
  __pyx_v_self->gap = 0;

  /* "_gappedbuf.pyx":29
 *         self.numelems = 0
 *         self.gap = 0
 *         self.gapsize = BUFSIZE             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->gapsize = 4096;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static void __pyx_pw_10_gappedbuf_12GappedBuffer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10_gappedbuf_12GappedBuffer_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10_gappedbuf_12GappedBuffer_2__dealloc__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":31
 *         self.gapsize = BUFSIZE
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.buf:
 *             free(self.buf)
 */

static void __pyx_pf_10_gappedbuf_12GappedBuffer_2__dealloc__(struct GappedBufferObj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "_gappedbuf.pyx":32
 * 
 *     def __dealloc__(self):
 *         if self.buf:             # <<<<<<<<<<<<<<
 *             free(self.buf)
 *             self.buf = NULL
 */
  __pyx_t_1 = (__pyx_v_self->buf != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":33
 *     def __dealloc__(self):
 *         if self.buf:
 *             free(self.buf)             # <<<<<<<<<<<<<<
 *             self.buf = NULL
 * 
 */
    free(__pyx_v_self->buf);

    /* "_gappedbuf.pyx":34
 *         if self.buf:
 *             free(self.buf)
 *             self.buf = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef void _resize_buf(self, Py_ssize_t size):
 */
    __pyx_v_self->buf = NULL;
    goto __pyx_L3;
  }
  __pyx_L3:;

  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":36
 *             self.buf = NULL
 * 
 *     cdef void _resize_buf(self, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *         cdef Py_UCS4 *buf = <Py_UCS4*>realloc(
 *                 self.buf, size * sizeof(Py_UCS4))
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__resize_buf(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_size) {
  Py_UCS4 *__pyx_v_buf;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_resize_buf", 0);

  /* "_gappedbuf.pyx":38
 *     cdef void _resize_buf(self, Py_ssize_t size):
 *         cdef Py_UCS4 *buf = <Py_UCS4*>realloc(
 *                 self.buf, size * sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 * 
 *         if not buf:
 */
  __pyx_v_buf = ((Py_UCS4 *)realloc(__pyx_v_self->buf, (__pyx_v_size * (sizeof(Py_UCS4)))));

  /* "_gappedbuf.pyx":40
 *                 self.buf, size * sizeof(Py_UCS4))
 * 
 *         if not buf:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self.buf = buf
 */
  __pyx_t_1 = ((!(__pyx_v_buf != 0)) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":41
 * 
 *         if not buf:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         self.buf = buf
 *         self.bufsize = size
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 41; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":42
 *         if not buf:
 *             raise MemoryError()
 *         self.buf = buf             # <<<<<<<<<<<<<<
 *         self.bufsize = size
 * 
 */
  __pyx_v_self->buf = __pyx_v_buf;

  /* "_gappedbuf.pyx":43
 *             raise MemoryError()
 *         self.buf = buf
 *         self.bufsize = size             # <<<<<<<<<<<<<<
 * 
 *     cdef void _expand_gap(self, Py_ssize_t size):
 */
  __pyx_v_self->bufsize = __pyx_v_size;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("_gappedbuf.GappedBuffer._resize_buf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":45
 *         self.bufsize = size
 * 
 *     cdef void _expand_gap(self, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *         if size <= self.gapsize:
 *             return
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__expand_gap(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_size) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_expand_gap", 0);

  /* "_gappedbuf.pyx":46
 * 
 *     cdef void _expand_gap(self, Py_ssize_t size):
 *         if size <= self.gapsize:             # <<<<<<<<<<<<<<
 *             return
 *         self._resize_buf(self.numelems+size)
 */
  __pyx_t_1 = ((__pyx_v_size <= __pyx_v_self->gapsize) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":47
 *     cdef void _expand_gap(self, Py_ssize_t size):
 *         if size <= self.gapsize:
 *             return             # <<<<<<<<<<<<<<
 *         self._resize_buf(self.numelems+size)
 * 
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":48
 *         if size <= self.gapsize:
 *             return
 *         self._resize_buf(self.numelems+size)             # <<<<<<<<<<<<<<
 * 
 *         if self.gap < self.numelems:
 */
  ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_resize_buf(__pyx_v_self, (__pyx_v_self->numelems + __pyx_v_size));

  /* "_gappedbuf.pyx":50
 *         self._resize_buf(self.numelems+size)
 * 
 *         if self.gap < self.numelems:             # <<<<<<<<<<<<<<
 *             # resize gap
 *             memmove(self.buf+self.gap+size, self.buf+self.gap+self.gapsize,
 */
  __pyx_t_1 = ((__pyx_v_self->gap < __pyx_v_self->numelems) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":53
 *             # resize gap
 *             memmove(self.buf+self.gap+size, self.buf+self.gap+self.gapsize,
 *                     (self.numelems-self.gap) * sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 * 
 *         self.gapsize = size
 */
    memmove(((__pyx_v_self->buf + __pyx_v_self->gap) + __pyx_v_size), ((__pyx_v_self->buf + __pyx_v_self->gap) + __pyx_v_self->gapsize), ((__pyx_v_self->numelems - __pyx_v_self->gap) * (sizeof(Py_UCS4))));
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":55
 *                     (self.numelems-self.gap) * sizeof(Py_UCS4))
 * 
 *         self.gapsize = size             # <<<<<<<<<<<<<<
 * 
 *     cdef void _shrink_gap(self):
 */
  __pyx_v_self->gapsize = __pyx_v_size;

  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":57
 *         self.gapsize = size
 * 
 *     cdef void _shrink_gap(self):             # <<<<<<<<<<<<<<
 *         DEF MAX_GAPSIZE = 4096
 *         if self.numelems + MAX_GAPSIZE > self.bufsize:
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__shrink_gap(struct GappedBufferObj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_shrink_gap", 0);

  /* "_gappedbuf.pyx":59
 *     cdef void _shrink_gap(self):
 *         DEF MAX_GAPSIZE = 4096
 *         if self.numelems + MAX_GAPSIZE > self.bufsize:             # <<<<<<<<<<<<<<
 *             return
 *         memmove(self.buf+self.gap+MAX_GAPSIZE,
 */
  __pyx_t_1 = (((__pyx_v_self->numelems + 4096) > __pyx_v_self->bufsize) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":60
 *         DEF MAX_GAPSIZE = 4096
 *         if self.numelems + MAX_GAPSIZE > self.bufsize:
 *             return             # <<<<<<<<<<<<<<
 *         memmove(self.buf+self.gap+MAX_GAPSIZE,
 *                 self.buf+self.gap+self.gapsize,
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":63
 *         memmove(self.buf+self.gap+MAX_GAPSIZE,
 *                 self.buf+self.gap+self.gapsize,
 *                 (self.numelems-self.gap) * sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 * 
 *         self.gapsize = MAX_GAPSIZE
 */
  memmove(((__pyx_v_self->buf + __pyx_v_self->gap) + 4096), ((__pyx_v_self->buf + __pyx_v_self->gap) + __pyx_v_self->gapsize), ((__pyx_v_self->numelems - __pyx_v_self->gap) * (sizeof(Py_UCS4))));

  /* "_gappedbuf.pyx":65
 *                 (self.numelems-self.gap) * sizeof(Py_UCS4))
 * 
 *         self.gapsize = MAX_GAPSIZE             # <<<<<<<<<<<<<<
 *         self._resize_buf(self.numelems+MAX_GAPSIZE)
 * 
 */
  __pyx_v_self->gapsize = 4096;

  /* "_gappedbuf.pyx":66
 * 
 *         self.gapsize = MAX_GAPSIZE
 *         self._resize_buf(self.numelems+MAX_GAPSIZE)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _move_gap(self, Py_ssize_t index):
 */
  ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_resize_buf(__pyx_v_self, (__pyx_v_self->numelems + 4096));

  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":68
 *         self._resize_buf(self.numelems+MAX_GAPSIZE)
 * 
 *     cdef void _move_gap(self, Py_ssize_t index):             # <<<<<<<<<<<<<<
 *         assert index <= self.bufsize
 *         if self.gap == index:
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__move_gap(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_move_gap", 0);

  /* "_gappedbuf.pyx":69
 * 
 *     cdef void _move_gap(self, Py_ssize_t index):
 *         assert index <= self.bufsize             # <<<<<<<<<<<<<<
 *         if self.gap == index:
 *             return
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!((__pyx_v_index <= __pyx_v_self->bufsize) != 0))) {
    PyErr_SetNone(PyExc_AssertionError);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  #endif

  /* "_gappedbuf.pyx":70
 *     cdef void _move_gap(self, Py_ssize_t index):
 *         assert index <= self.bufsize
 *         if self.gap == index:             # <<<<<<<<<<<<<<
 *             return
 *         if self.gapsize:
 */
  __pyx_t_1 = ((__pyx_v_self->gap == __pyx_v_index) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":71
 *         assert index <= self.bufsize
 *         if self.gap == index:
 *             return             # <<<<<<<<<<<<<<
 *         if self.gapsize:
 *             if index < self.gap:
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":72
 *         if self.gap == index:
 *             return
 *         if self.gapsize:             # <<<<<<<<<<<<<<
 *             if index < self.gap:
 *                 memmove(self.buf+index+self.gapsize, self.buf+index,
 */
  __pyx_t_1 = (__pyx_v_self->gapsize != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":73
 *             return
 *         if self.gapsize:
 *             if index < self.gap:             # <<<<<<<<<<<<<<
 *                 memmove(self.buf+index+self.gapsize, self.buf+index,
 *                         (self.gap - index)*sizeof(Py_UCS4))
 */
    __pyx_t_1 = ((__pyx_v_index < __pyx_v_self->gap) != 0);
    if (__pyx_t_1) {

      /* "_gappedbuf.pyx":75
 *             if index < self.gap:
 *                 memmove(self.buf+index+self.gapsize, self.buf+index,
 *                         (self.gap - index)*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *             else:
 *                 memmove(self.buf+self.gap, self.buf+self.gap+self.gapsize,
 */
      memmove(((__pyx_v_self->buf + __pyx_v_index) + __pyx_v_self->gapsize), (__pyx_v_self->buf + __pyx_v_index), ((__pyx_v_self->gap - __pyx_v_index) * (sizeof(Py_UCS4))));
      goto __pyx_L5;
    }
    /*else*/ {

      /* "_gappedbuf.pyx":78
 *             else:
 *                 memmove(self.buf+self.gap, self.buf+self.gap+self.gapsize,
 *                         (index - self.gap)*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *         self.gap = index
 * 
 */
      memmove((__pyx_v_self->buf + __pyx_v_self->gap), ((__pyx_v_self->buf + __pyx_v_self->gap) + __pyx_v_self->gapsize), ((__pyx_v_index - __pyx_v_self->gap) * (sizeof(Py_UCS4))));
    }
    __pyx_L5:;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":79
 *                 memmove(self.buf+self.gap, self.buf+self.gap+self.gapsize,
 *                         (index - self.gap)*sizeof(Py_UCS4))
 *         self.gap = index             # <<<<<<<<<<<<<<
 * 
 *     cdef _insert(self, Py_ssize_t index, unicode s):
 */
  __pyx_v_self->gap = __pyx_v_index;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("_gappedbuf.GappedBuffer._move_gap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "_gappedbuf.pyx":81
 *         self.gap = index
 * 
 *     cdef _insert(self, Py_ssize_t index, unicode s):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t size = len(s)
 *         if not (0 <= index <= self.numelems):
 */

static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer__insert(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_insert", 0);

  /* "_gappedbuf.pyx":82
 * 
 *     cdef _insert(self, Py_ssize_t index, unicode s):
 *         cdef Py_ssize_t size = len(s)             # <<<<<<<<<<<<<<
 *         if not (0 <= index <= self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 */
  if (unlikely(((PyObject *)__pyx_v_s) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = __Pyx_PyUnicode_GET_LENGTH(((PyObject *)__pyx_v_s)); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_size = __pyx_t_1;

  /* "_gappedbuf.pyx":83
 *     cdef _insert(self, Py_ssize_t index, unicode s):
 *         cdef Py_ssize_t size = len(s)
 *         if not (0 <= index <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 index, self.numelems))
 */
  __pyx_t_2 = (0 <= __pyx_v_index);
  if (__pyx_t_2) {
    __pyx_t_2 = (__pyx_v_index <= __pyx_v_self->numelems);
  }
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "_gappedbuf.pyx":85
 *         if not (0 <= index <= self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 index, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         if size:
 */
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_index); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 85; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 85; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 85; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_1), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 84; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 84; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 84; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 84; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":87
 *                                 index, self.numelems))
 * 
 *         if size:             # <<<<<<<<<<<<<<
 *             self._move_gap(index)
 *             if self.gapsize < size:
 */
  __pyx_t_3 = (__pyx_v_size != 0);
  if (__pyx_t_3) {

    /* "_gappedbuf.pyx":88
 * 
 *         if size:
 *             self._move_gap(index)             # <<<<<<<<<<<<<<
 *             if self.gapsize < size:
 *                 self._expand_gap(max(1024, size*2))
 */
    ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_move_gap(__pyx_v_self, __pyx_v_index);

    /* "_gappedbuf.pyx":89
 *         if size:
 *             self._move_gap(index)
 *             if self.gapsize < size:             # <<<<<<<<<<<<<<
 *                 self._expand_gap(max(1024, size*2))
 *             rtdef.PyUnicode_AsUCS4(s, self.buf+index, size, False)
 */
    __pyx_t_3 = ((__pyx_v_self->gapsize < __pyx_v_size) != 0);
    if (__pyx_t_3) {

      /* "_gappedbuf.pyx":90
 *             self._move_gap(index)
 *             if self.gapsize < size:
 *                 self._expand_gap(max(1024, size*2))             # <<<<<<<<<<<<<<
 *             rtdef.PyUnicode_AsUCS4(s, self.buf+index, size, False)
 * 
 */
      __pyx_t_1 = (__pyx_v_size * 2);
      __pyx_t_7 = 1024;
      if (((__pyx_t_1 > __pyx_t_7) != 0)) {
        __pyx_t_8 = __pyx_t_1;
      } else {
        __pyx_t_8 = __pyx_t_7;
      }
      ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_expand_gap(__pyx_v_self, __pyx_t_8);
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "_gappedbuf.pyx":91
 *             if self.gapsize < size:
 *                 self._expand_gap(max(1024, size*2))
 *             rtdef.PyUnicode_AsUCS4(s, self.buf+index, size, False)             # <<<<<<<<<<<<<<
 * 
 *             self.numelems += size
 */
    __pyx_t_9 = PyUnicode_AsUCS4(((PyObject *)__pyx_v_s), (__pyx_v_self->buf + __pyx_v_index), __pyx_v_size, 0); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "_gappedbuf.pyx":93
 *             rtdef.PyUnicode_AsUCS4(s, self.buf+index, size, False)
 * 
 *             self.numelems += size             # <<<<<<<<<<<<<<
 *             self.gap += size
 *             self.gapsize -= size
 */
    __pyx_v_self->numelems = (__pyx_v_self->numelems + __pyx_v_size);

    /* "_gappedbuf.pyx":94
 * 
 *             self.numelems += size
 *             self.gap += size             # <<<<<<<<<<<<<<
 *             self.gapsize -= size
 * 
 */
    __pyx_v_self->gap = (__pyx_v_self->gap + __pyx_v_size);

    /* "_gappedbuf.pyx":95
 *             self.numelems += size
 *             self.gap += size
 *             self.gapsize -= size             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, Py_ssize_t index, unicode s):
 */
    __pyx_v_self->gapsize = (__pyx_v_self->gapsize - __pyx_v_size);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer._insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":97
 *             self.gapsize -= size
 * 
 *     cpdef insert(self, Py_ssize_t index, unicode s):             # <<<<<<<<<<<<<<
 *         self._insert(index, s)
 * 
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_5insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_insert(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__insert); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_5insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_index); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_s));
      PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_s));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_s));
      __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":98
 * 
 *     cpdef insert(self, Py_ssize_t index, unicode s):
 *         self._insert(index, s)             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, unicode s):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_insert(__pyx_v_self, __pyx_v_index, __pyx_v_s); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 98; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_5insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_5insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_v_s = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__index,&__pyx_n_s__s,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_index = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_index == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_s = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyUnicode_Type), 1, "s", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_4insert(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_index, __pyx_v_s);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":97
 *             self.gapsize -= size
 * 
 *     cpdef insert(self, Py_ssize_t index, unicode s):             # <<<<<<<<<<<<<<
 *         self._insert(index, s)
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_4insert(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->insert(__pyx_v_self, __pyx_v_index, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":100
 *         self._insert(index, s)
 * 
 *     cpdef append(self, unicode s):             # <<<<<<<<<<<<<<
 *         self.insert(self.numelems, s)
 * 
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7append(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_append(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__append); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_7append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_s));
      PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_s));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_s));
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":101
 * 
 *     cpdef append(self, unicode s):
 *         self.insert(self.numelems, s)             # <<<<<<<<<<<<<<
 * 
 *     cdef _delete(self, Py_ssize_t begin, Py_ssize_t end):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->insert(__pyx_v_self, __pyx_v_self->numelems, __pyx_v_s, 0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7append(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7append(PyObject *__pyx_v_self, PyObject *__pyx_v_s) {
  CYTHON_UNUSED int __pyx_lineno = 0;
  CYTHON_UNUSED const char *__pyx_filename = NULL;
  CYTHON_UNUSED int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyUnicode_Type), 1, "s", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_6append(((struct GappedBufferObj *)__pyx_v_self), ((PyObject*)__pyx_v_s));
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":100
 *         self._insert(index, s)
 * 
 *     cpdef append(self, unicode s):             # <<<<<<<<<<<<<<
 *         self.insert(self.numelems, s)
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_6append(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->append(__pyx_v_self, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":103
 *         self.insert(self.numelems, s)
 * 
 *     cdef _delete(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 */

static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer__delete(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_delete", 0);

  /* "_gappedbuf.pyx":104
 * 
 *     cdef _delete(self, Py_ssize_t begin, Py_ssize_t end):
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_1 = (0 <= __pyx_v_begin);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":106
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t size = end-begin
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":108
 *                                 begin, end, self.numelems))
 * 
 *         cdef Py_ssize_t size = end-begin             # <<<<<<<<<<<<<<
 *         if not size:
 *             return
 */
  __pyx_v_size = (__pyx_v_end - __pyx_v_begin);

  /* "_gappedbuf.pyx":109
 * 
 *         cdef Py_ssize_t size = end-begin
 *         if not size:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_size != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":110
 *         cdef Py_ssize_t size = end-begin
 *         if not size:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._move_gap(begin)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":112
 *             return
 * 
 *         self._move_gap(begin)             # <<<<<<<<<<<<<<
 *         self.gapsize += size
 *         self.numelems -= size
 */
  ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_move_gap(__pyx_v_self, __pyx_v_begin);

  /* "_gappedbuf.pyx":113
 * 
 *         self._move_gap(begin)
 *         self.gapsize += size             # <<<<<<<<<<<<<<
 *         self.numelems -= size
 *         self._shrink_gap()
 */
  __pyx_v_self->gapsize = (__pyx_v_self->gapsize + __pyx_v_size);

  /* "_gappedbuf.pyx":114
 *         self._move_gap(begin)
 *         self.gapsize += size
 *         self.numelems -= size             # <<<<<<<<<<<<<<
 *         self._shrink_gap()
 * 
 */
  __pyx_v_self->numelems = (__pyx_v_self->numelems - __pyx_v_size);

  /* "_gappedbuf.pyx":115
 *         self.gapsize += size
 *         self.numelems -= size
 *         self._shrink_gap()             # <<<<<<<<<<<<<<
 * 
 *     cpdef delete(self, Py_ssize_t begin, Py_ssize_t end):
 */
  ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_shrink_gap(__pyx_v_self);

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer._delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":117
 *         self._shrink_gap()
 * 
 *     cpdef delete(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 * 
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_9delete(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_delete(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delete", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__delete); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_9delete)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":118
 * 
 *     cpdef delete(self, Py_ssize_t begin, Py_ssize_t end):
 *         self._delete(begin, end)             # <<<<<<<<<<<<<<
 * 
 *     cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_delete(__pyx_v_self, __pyx_v_begin, __pyx_v_end); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_9delete(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_9delete(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("delete (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("delete", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "delete") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("delete", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_8delete(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":117
 *         self._shrink_gap()
 * 
 *     cpdef delete(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_8delete(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delete", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->delete(__pyx_v_self, __pyx_v_begin, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":120
 *         self._delete(begin, end)
 * 
 *     cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 *         self._insert(begin, s)
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_11replace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_replace(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__replace); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_11replace)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_INCREF(((PyObject *)__pyx_v_s));
      PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_v_s));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_s));
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":121
 * 
 *     cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):
 *         self._delete(begin, end)             # <<<<<<<<<<<<<<
 *         self._insert(begin, s)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_delete(__pyx_v_self, __pyx_v_begin, __pyx_v_end); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 121; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_gappedbuf.pyx":122
 *     cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):
 *         self._delete(begin, end)
 *         self._insert(begin, s)             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_UCS4 _get_max(self, Py_ssize_t begin, Py_ssize_t end) nogil:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_insert(__pyx_v_self, __pyx_v_begin, __pyx_v_s); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 122; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_11replace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_11replace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  PyObject *__pyx_v_s = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,&__pyx_n_s__s,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("replace", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("replace", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "replace") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_s = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("replace", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyUnicode_Type), 1, "s", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_10replace(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end, __pyx_v_s);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":120
 *         self._delete(begin, end)
 * 
 *     cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 *         self._insert(begin, s)
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_10replace(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->replace(__pyx_v_self, __pyx_v_begin, __pyx_v_end, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":124
 *         self._insert(begin, s)
 * 
 *     cdef Py_UCS4 _get_max(self, Py_ssize_t begin, Py_ssize_t end) nogil:             # <<<<<<<<<<<<<<
 *         cdef Py_UCS4 maxchar = 0
 *         cdef Py_ssize_t b, e
 */

static Py_UCS4 __pyx_f_10_gappedbuf_12GappedBuffer__get_max(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_UCS4 __pyx_v_maxchar;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  Py_UCS4 __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  Py_UCS4 __pyx_t_6;

  /* "_gappedbuf.pyx":125
 * 
 *     cdef Py_UCS4 _get_max(self, Py_ssize_t begin, Py_ssize_t end) nogil:
 *         cdef Py_UCS4 maxchar = 0             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t b, e
 * 
 */
  __pyx_v_maxchar = 0;

  /* "_gappedbuf.pyx":128
 *         cdef Py_ssize_t b, e
 * 
 *         e = min(end, self.gap)             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t p
 */
  __pyx_t_1 = __pyx_v_self->gap;
  __pyx_t_2 = __pyx_v_end;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_e = __pyx_t_3;

  /* "_gappedbuf.pyx":131
 * 
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *             maxchar = max(maxchar, self.buf[p])
 *         b = max(begin, self.gap)
 */
  __pyx_t_3 = __pyx_v_e;
  for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_3; __pyx_v_p++) {

    /* "_gappedbuf.pyx":132
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:
 *             maxchar = max(maxchar, self.buf[p])             # <<<<<<<<<<<<<<
 *         b = max(begin, self.gap)
 *         for p from b <= p < end:
 */
    __pyx_t_4 = (__pyx_v_self->buf[__pyx_v_p]);
    __pyx_t_5 = __pyx_v_maxchar;
    if (((__pyx_t_4 > __pyx_t_5) != 0)) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_v_maxchar = __pyx_t_6;
  }

  /* "_gappedbuf.pyx":133
 *         for p from begin <= p < e:
 *             maxchar = max(maxchar, self.buf[p])
 *         b = max(begin, self.gap)             # <<<<<<<<<<<<<<
 *         for p from b <= p < end:
 *             maxchar = max(maxchar, self.buf[p+self.gapsize])
 */
  __pyx_t_3 = __pyx_v_self->gap;
  __pyx_t_1 = __pyx_v_begin;
  if (((__pyx_t_3 > __pyx_t_1) != 0)) {
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  __pyx_v_b = __pyx_t_2;

  /* "_gappedbuf.pyx":134
 *             maxchar = max(maxchar, self.buf[p])
 *         b = max(begin, self.gap)
 *         for p from b <= p < end:             # <<<<<<<<<<<<<<
 *             maxchar = max(maxchar, self.buf[p+self.gapsize])
 *         return maxchar
 */
  __pyx_t_2 = __pyx_v_end;
  for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_2; __pyx_v_p++) {

    /* "_gappedbuf.pyx":135
 *         b = max(begin, self.gap)
 *         for p from b <= p < end:
 *             maxchar = max(maxchar, self.buf[p+self.gapsize])             # <<<<<<<<<<<<<<
 *         return maxchar
 * 
 */
    __pyx_t_6 = (__pyx_v_self->buf[(__pyx_v_p + __pyx_v_self->gapsize)]);
    __pyx_t_4 = __pyx_v_maxchar;
    if (((__pyx_t_6 > __pyx_t_4) != 0)) {
      __pyx_t_5 = __pyx_t_6;
    } else {
      __pyx_t_5 = __pyx_t_4;
    }
    __pyx_v_maxchar = __pyx_t_5;
  }

  /* "_gappedbuf.pyx":136
 *         for p from b <= p < end:
 *             maxchar = max(maxchar, self.buf[p+self.gapsize])
 *         return maxchar             # <<<<<<<<<<<<<<
 * 
 *     cdef void _copy_chars1(self, void *dest, Py_ssize_t begin,
 */
  __pyx_r = __pyx_v_maxchar;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "_gappedbuf.pyx":138
 *         return maxchar
 * 
 *     cdef void _copy_chars1(self, void *dest, Py_ssize_t begin,             # <<<<<<<<<<<<<<
 *             Py_ssize_t end) nogil:
 *         cdef rtdef.Py_UCS1 *buf = <rtdef.Py_UCS1 *>dest
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__copy_chars1(struct GappedBufferObj *__pyx_v_self, void *__pyx_v_dest, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_UCS1 *__pyx_v_buf;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;

  /* "_gappedbuf.pyx":140
 *     cdef void _copy_chars1(self, void *dest, Py_ssize_t begin,
 *             Py_ssize_t end) nogil:
 *         cdef rtdef.Py_UCS1 *buf = <rtdef.Py_UCS1 *>dest             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t b, e
 * 
 */
  __pyx_v_buf = ((Py_UCS1 *)__pyx_v_dest);

  /* "_gappedbuf.pyx":144
 * 
 *         # copy before gap
 *         e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:
 */
  __pyx_t_1 = __pyx_v_self->gap;
  __pyx_t_2 = __pyx_v_end;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_e = __pyx_t_3;

  /* "_gappedbuf.pyx":146
 *         e = min(end, self.gap)
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *             buf[0] = self.buf[p]
 *             buf += 1
 */
  __pyx_t_3 = __pyx_v_e;
  for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_3; __pyx_v_p++) {

    /* "_gappedbuf.pyx":147
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:
 *             buf[0] = self.buf[p]             # <<<<<<<<<<<<<<
 *             buf += 1
 * 
 */
    (__pyx_v_buf[0]) = (__pyx_v_self->buf[__pyx_v_p]);

    /* "_gappedbuf.pyx":148
 *         for p from begin <= p < e:
 *             buf[0] = self.buf[p]
 *             buf += 1             # <<<<<<<<<<<<<<
 * 
 *         # copy after gap
 */
    __pyx_v_buf = (__pyx_v_buf + 1);
  }

  /* "_gappedbuf.pyx":151
 * 
 *         # copy after gap
 *         b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]
 */
  __pyx_t_3 = __pyx_v_self->gap;
  __pyx_t_1 = __pyx_v_begin;
  if (((__pyx_t_3 > __pyx_t_1) != 0)) {
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  __pyx_v_b = (__pyx_t_2 + __pyx_v_self->gapsize);

  /* "_gappedbuf.pyx":152
 *         # copy after gap
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *             buf[0] = self.buf[p]
 *             buf += 1
 */
  __pyx_t_2 = (__pyx_v_end + __pyx_v_self->gapsize);
  for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_2; __pyx_v_p++) {

    /* "_gappedbuf.pyx":153
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]             # <<<<<<<<<<<<<<
 *             buf += 1
 * 
 */
    (__pyx_v_buf[0]) = (__pyx_v_self->buf[__pyx_v_p]);

    /* "_gappedbuf.pyx":154
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]
 *             buf += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void _copy_chars2(self, void *dest, Py_ssize_t begin,
 */
    __pyx_v_buf = (__pyx_v_buf + 1);
  }

}

/* "_gappedbuf.pyx":156
 *             buf += 1
 * 
 *     cdef void _copy_chars2(self, void *dest, Py_ssize_t begin,             # <<<<<<<<<<<<<<
 *             Py_ssize_t end) nogil:
 *         cdef rtdef.Py_UCS2 *buf = <rtdef.Py_UCS2 *>dest
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__copy_chars2(struct GappedBufferObj *__pyx_v_self, void *__pyx_v_dest, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_UCS2 *__pyx_v_buf;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;

  /* "_gappedbuf.pyx":158
 *     cdef void _copy_chars2(self, void *dest, Py_ssize_t begin,
 *             Py_ssize_t end) nogil:
 *         cdef rtdef.Py_UCS2 *buf = <rtdef.Py_UCS2 *>dest             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t b, e
 * 
 */
  __pyx_v_buf = ((Py_UCS2 *)__pyx_v_dest);

  /* "_gappedbuf.pyx":162
 * 
 *         # copy before gap
 *         e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:
 */
  __pyx_t_1 = __pyx_v_self->gap;
  __pyx_t_2 = __pyx_v_end;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_e = __pyx_t_3;

  /* "_gappedbuf.pyx":164
 *         e = min(end, self.gap)
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *             buf[0] = self.buf[p]
 *             buf += 1
 */
  __pyx_t_3 = __pyx_v_e;
  for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_3; __pyx_v_p++) {

    /* "_gappedbuf.pyx":165
 *         cdef Py_ssize_t p
 *         for p from begin <= p < e:
 *             buf[0] = self.buf[p]             # <<<<<<<<<<<<<<
 *             buf += 1
 * 
 */
    (__pyx_v_buf[0]) = (__pyx_v_self->buf[__pyx_v_p]);

    /* "_gappedbuf.pyx":166
 *         for p from begin <= p < e:
 *             buf[0] = self.buf[p]
 *             buf += 1             # <<<<<<<<<<<<<<
 * 
 *         # copy after gap
 */
    __pyx_v_buf = (__pyx_v_buf + 1);
  }

  /* "_gappedbuf.pyx":169
 * 
 *         # copy after gap
 *         b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]
 */
  __pyx_t_3 = __pyx_v_self->gap;
  __pyx_t_1 = __pyx_v_begin;
  if (((__pyx_t_3 > __pyx_t_1) != 0)) {
    __pyx_t_2 = __pyx_t_3;
  } else {
    __pyx_t_2 = __pyx_t_1;
  }
  __pyx_v_b = (__pyx_t_2 + __pyx_v_self->gapsize);

  /* "_gappedbuf.pyx":170
 *         # copy after gap
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *             buf[0] = self.buf[p]
 *             buf += 1
 */
  __pyx_t_2 = (__pyx_v_end + __pyx_v_self->gapsize);
  for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_2; __pyx_v_p++) {

    /* "_gappedbuf.pyx":171
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]             # <<<<<<<<<<<<<<
 *             buf += 1
 * 
 */
    (__pyx_v_buf[0]) = (__pyx_v_self->buf[__pyx_v_p]);

    /* "_gappedbuf.pyx":172
 *         for p from b <= p < end+self.gapsize:
 *             buf[0] = self.buf[p]
 *             buf += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void _copy_chars4(self, void *dest, Py_ssize_t begin,
 */
    __pyx_v_buf = (__pyx_v_buf + 1);
  }

}

/* "_gappedbuf.pyx":174
 *             buf += 1
 * 
 *     cdef void _copy_chars4(self, void *dest, Py_ssize_t begin,             # <<<<<<<<<<<<<<
 *             Py_ssize_t end) nogil:
 *         cdef Py_ssize_t n = end - begin
 */

static void __pyx_f_10_gappedbuf_12GappedBuffer__copy_chars4(struct GappedBufferObj *__pyx_v_self, void *__pyx_v_dest, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_ssize_t __pyx_v_n;
  Py_ssize_t __pyx_v_former;
  int __pyx_t_1;

  /* "_gappedbuf.pyx":176
 *     cdef void _copy_chars4(self, void *dest, Py_ssize_t begin,
 *             Py_ssize_t end) nogil:
 *         cdef Py_ssize_t n = end - begin             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t former
 * 
 */
  __pyx_v_n = (__pyx_v_end - __pyx_v_begin);

  /* "_gappedbuf.pyx":179
 *         cdef Py_ssize_t former
 * 
 *         if self.gap < begin:             # <<<<<<<<<<<<<<
 *             memcpy(dest, self.buf+begin+self.gapsize, n*sizeof(Py_UCS4))
 *         elif end <= self.gap:
 */
  __pyx_t_1 = ((__pyx_v_self->gap < __pyx_v_begin) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":180
 * 
 *         if self.gap < begin:
 *             memcpy(dest, self.buf+begin+self.gapsize, n*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *         elif end <= self.gap:
 *             memcpy(dest, self.buf+begin, n*sizeof(Py_UCS4))
 */
    memcpy(__pyx_v_dest, ((__pyx_v_self->buf + __pyx_v_begin) + __pyx_v_self->gapsize), (__pyx_v_n * (sizeof(Py_UCS4))));
    goto __pyx_L3;
  }

  /* "_gappedbuf.pyx":181
 *         if self.gap < begin:
 *             memcpy(dest, self.buf+begin+self.gapsize, n*sizeof(Py_UCS4))
 *         elif end <= self.gap:             # <<<<<<<<<<<<<<
 *             memcpy(dest, self.buf+begin, n*sizeof(Py_UCS4))
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_end <= __pyx_v_self->gap) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":182
 *             memcpy(dest, self.buf+begin+self.gapsize, n*sizeof(Py_UCS4))
 *         elif end <= self.gap:
 *             memcpy(dest, self.buf+begin, n*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *         else:
 *             former = self.gap - begin
 */
    memcpy(__pyx_v_dest, (__pyx_v_self->buf + __pyx_v_begin), (__pyx_v_n * (sizeof(Py_UCS4))));
    goto __pyx_L3;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":184
 *             memcpy(dest, self.buf+begin, n*sizeof(Py_UCS4))
 *         else:
 *             former = self.gap - begin             # <<<<<<<<<<<<<<
 *             memcpy(dest, self.buf+begin, former*sizeof(Py_UCS4))
 *             memcpy(dest+former*sizeof(Py_UCS4), self.buf+self.gap+self.gapsize,
 */
    __pyx_v_former = (__pyx_v_self->gap - __pyx_v_begin);

    /* "_gappedbuf.pyx":185
 *         else:
 *             former = self.gap - begin
 *             memcpy(dest, self.buf+begin, former*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 *             memcpy(dest+former*sizeof(Py_UCS4), self.buf+self.gap+self.gapsize,
 *                     (n-former)*sizeof(Py_UCS4))
 */
    memcpy(__pyx_v_dest, (__pyx_v_self->buf + __pyx_v_begin), (__pyx_v_former * (sizeof(Py_UCS4))));

    /* "_gappedbuf.pyx":187
 *             memcpy(dest, self.buf+begin, former*sizeof(Py_UCS4))
 *             memcpy(dest+former*sizeof(Py_UCS4), self.buf+self.gap+self.gapsize,
 *                     (n-former)*sizeof(Py_UCS4))             # <<<<<<<<<<<<<<
 * 
 *     cpdef get(self, Py_ssize_t begin, Py_ssize_t end):
 */
    memcpy((__pyx_v_dest + (__pyx_v_former * (sizeof(Py_UCS4)))), ((__pyx_v_self->buf + __pyx_v_self->gap) + __pyx_v_self->gapsize), ((__pyx_v_n - __pyx_v_former) * (sizeof(Py_UCS4))));
  }
  __pyx_L3:;

}

/* "_gappedbuf.pyx":189
 *                     (n-former)*sizeof(Py_UCS4))
 * 
 *     cpdef get(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise IndexError('Invalid range: %d, %d, %d' % (
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_13get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_get(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_n;
  Py_UCS4 __pyx_v_maxchar;
  PyObject *__pyx_v_ret = 0;
  void *__pyx_v_buf;
  PyObject *__pyx_v_nbytes = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__get); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_13get)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":190
 * 
 *     cpdef get(self, Py_ssize_t begin, Py_ssize_t end):
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise IndexError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_5 = (0 <= __pyx_v_begin);
  if (__pyx_t_5) {
    __pyx_t_5 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_5) {
      __pyx_t_5 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
  if (__pyx_t_6) {

    /* "_gappedbuf.pyx":192
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise IndexError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t n = end - begin
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_4));
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_4));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Call(__pyx_builtin_IndexError, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":194
 *                                 begin, end, self.numelems))
 * 
 *         cdef Py_ssize_t n = end - begin             # <<<<<<<<<<<<<<
 *         cdef Py_UCS4 maxchar = self._get_max(begin, end)
 *         cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)
 */
  __pyx_v_n = (__pyx_v_end - __pyx_v_begin);

  /* "_gappedbuf.pyx":195
 * 
 *         cdef Py_ssize_t n = end - begin
 *         cdef Py_UCS4 maxchar = self._get_max(begin, end)             # <<<<<<<<<<<<<<
 *         cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)
 *         cdef void *buf = rtdef.PyUnicode_DATA(ret)
 */
  __pyx_v_maxchar = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_get_max(__pyx_v_self, __pyx_v_begin, __pyx_v_end);

  /* "_gappedbuf.pyx":196
 *         cdef Py_ssize_t n = end - begin
 *         cdef Py_UCS4 maxchar = self._get_max(begin, end)
 *         cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)             # <<<<<<<<<<<<<<
 *         cdef void *buf = rtdef.PyUnicode_DATA(ret)
 *         cdef nbytes = rtdef.PyUnicode_KIND(ret)
 */
  __pyx_t_4 = PyUnicode_New(__pyx_v_n, __pyx_v_maxchar); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected unicode, got %.200s", Py_TYPE(__pyx_t_4)->tp_name), 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_ret = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "_gappedbuf.pyx":197
 *         cdef Py_UCS4 maxchar = self._get_max(begin, end)
 *         cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)
 *         cdef void *buf = rtdef.PyUnicode_DATA(ret)             # <<<<<<<<<<<<<<
 *         cdef nbytes = rtdef.PyUnicode_KIND(ret)
 * 
 */
  __pyx_t_7 = PyUnicode_DATA(((PyObject *)__pyx_v_ret)); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 197; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_buf = __pyx_t_7;

  /* "_gappedbuf.pyx":198
 *         cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)
 *         cdef void *buf = rtdef.PyUnicode_DATA(ret)
 *         cdef nbytes = rtdef.PyUnicode_KIND(ret)             # <<<<<<<<<<<<<<
 * 
 *         if nbytes == 1:
 */
  __pyx_t_8 = PyUnicode_KIND(((PyObject *)__pyx_v_ret)); if (unlikely(__pyx_t_8 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = PyInt_FromLong(__pyx_t_8); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_nbytes = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_gappedbuf.pyx":200
 *         cdef nbytes = rtdef.PyUnicode_KIND(ret)
 * 
 *         if nbytes == 1:             # <<<<<<<<<<<<<<
 *             self._copy_chars1(buf, begin, end)
 *         elif nbytes == 2:
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_nbytes, __pyx_int_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "_gappedbuf.pyx":201
 * 
 *         if nbytes == 1:
 *             self._copy_chars1(buf, begin, end)             # <<<<<<<<<<<<<<
 *         elif nbytes == 2:
 *             self._copy_chars2(buf, begin, end)
 */
    ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_copy_chars1(__pyx_v_self, __pyx_v_buf, __pyx_v_begin, __pyx_v_end);
    goto __pyx_L4;
  }

  /* "_gappedbuf.pyx":202
 *         if nbytes == 1:
 *             self._copy_chars1(buf, begin, end)
 *         elif nbytes == 2:             # <<<<<<<<<<<<<<
 *             self._copy_chars2(buf, begin, end)
 *         elif nbytes == 4:
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_nbytes, __pyx_int_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "_gappedbuf.pyx":203
 *             self._copy_chars1(buf, begin, end)
 *         elif nbytes == 2:
 *             self._copy_chars2(buf, begin, end)             # <<<<<<<<<<<<<<
 *         elif nbytes == 4:
 *             self._copy_chars4(buf, begin, end)
 */
    ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_copy_chars2(__pyx_v_self, __pyx_v_buf, __pyx_v_begin, __pyx_v_end);
    goto __pyx_L4;
  }

  /* "_gappedbuf.pyx":204
 *         elif nbytes == 2:
 *             self._copy_chars2(buf, begin, end)
 *         elif nbytes == 4:             # <<<<<<<<<<<<<<
 *             self._copy_chars4(buf, begin, end)
 *         else:
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_nbytes, __pyx_int_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "_gappedbuf.pyx":205
 *             self._copy_chars2(buf, begin, end)
 *         elif nbytes == 4:
 *             self._copy_chars4(buf, begin, end)             # <<<<<<<<<<<<<<
 *         else:
 *             raise RuntimeError()
 */
    ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_copy_chars4(__pyx_v_self, __pyx_v_buf, __pyx_v_begin, __pyx_v_end);
    goto __pyx_L4;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":207
 *             self._copy_chars4(buf, begin, end)
 *         else:
 *             raise RuntimeError()             # <<<<<<<<<<<<<<
 * 
 *         return ret
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 207; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 207; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":209
 *             raise RuntimeError()
 * 
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_ret));
  __pyx_r = ((PyObject *)__pyx_v_ret);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_nbytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_13get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_13get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_12get(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":189
 *                     (n-former)*sizeof(Py_UCS4))
 * 
 *     cpdef get(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise IndexError('Invalid range: %d, %d, %d' % (
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_12get(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_begin, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static Py_ssize_t __pyx_pw_10_gappedbuf_12GappedBuffer_15__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10_gappedbuf_12GappedBuffer_15__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_14__len__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":211
 *         return ret
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.numelems
 * 
 */

static Py_ssize_t __pyx_pf_10_gappedbuf_12GappedBuffer_14__len__(struct GappedBufferObj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "_gappedbuf.pyx":212
 * 
 *     def __len__(self):
 *         return self.numelems             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_slice(self, object obj,
 */
  __pyx_r = __pyx_v_self->numelems;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":214
 *         return self.numelems
 * 
 *     cdef int _get_slice(self, object obj,             # <<<<<<<<<<<<<<
 *             Py_ssize_t *f, Py_ssize_t *t) except *:
 *         cdef Py_ssize_t step, slicelength
 */

static int __pyx_f_10_gappedbuf_12GappedBuffer__get_slice(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_obj, Py_ssize_t *__pyx_v_f, Py_ssize_t *__pyx_v_t) {
  Py_ssize_t __pyx_v_step;
  Py_ssize_t __pyx_v_slicelength;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_slice", 0);

  /* "_gappedbuf.pyx":218
 *         cdef Py_ssize_t step, slicelength
 * 
 *         if rtdef.PySlice_Check(obj):             # <<<<<<<<<<<<<<
 *             rtdef.PySlice_GetIndicesEx(obj, self.numelems, f, t,
 *                      &step, &slicelength)
 */
  __pyx_t_1 = (PySlice_Check(__pyx_v_obj) != 0);
  if (__pyx_t_1) {

    /* "_gappedbuf.pyx":220
 *         if rtdef.PySlice_Check(obj):
 *             rtdef.PySlice_GetIndicesEx(obj, self.numelems, f, t,
 *                      &step, &slicelength)             # <<<<<<<<<<<<<<
 *             if step != 1:
 *                 raise ValueError('step is not supported')
 */
    __pyx_t_2 = PySlice_GetIndicesEx(__pyx_v_obj, __pyx_v_self->numelems, __pyx_v_f, __pyx_v_t, (&__pyx_v_step), (&__pyx_v_slicelength)); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "_gappedbuf.pyx":221
 *             rtdef.PySlice_GetIndicesEx(obj, self.numelems, f, t,
 *                      &step, &slicelength)
 *             if step != 1:             # <<<<<<<<<<<<<<
 *                 raise ValueError('step is not supported')
 *             return 1
 */
    __pyx_t_1 = ((__pyx_v_step != 1) != 0);
    if (__pyx_t_1) {

      /* "_gappedbuf.pyx":222
 *                      &step, &slicelength)
 *             if step != 1:
 *                 raise ValueError('step is not supported')             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
      __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L4;
    }
    __pyx_L4:;

    /* "_gappedbuf.pyx":223
 *             if step != 1:
 *                 raise ValueError('step is not supported')
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             f[0] = obj
 */
    __pyx_r = 1;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":225
 *             return 1
 *         else:
 *             f[0] = obj             # <<<<<<<<<<<<<<
 *             if f[0] < 0:
 *                 f[0] += self.numelems
 */
    __pyx_t_4 = __Pyx_PyIndex_AsSsize_t(__pyx_v_obj); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    (__pyx_v_f[0]) = __pyx_t_4;

    /* "_gappedbuf.pyx":226
 *         else:
 *             f[0] = obj
 *             if f[0] < 0:             # <<<<<<<<<<<<<<
 *                 f[0] += self.numelems
 *             if not (0 <= f[0] < self.numelems):
 */
    __pyx_t_1 = (((__pyx_v_f[0]) < 0) != 0);
    if (__pyx_t_1) {

      /* "_gappedbuf.pyx":227
 *             f[0] = obj
 *             if f[0] < 0:
 *                 f[0] += self.numelems             # <<<<<<<<<<<<<<
 *             if not (0 <= f[0] < self.numelems):
 *                 raise IndexError('Invalid index value: %d, %d' % (
 */
      __pyx_t_5 = 0;
      (__pyx_v_f[__pyx_t_5]) = ((__pyx_v_f[__pyx_t_5]) + __pyx_v_self->numelems);
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "_gappedbuf.pyx":228
 *             if f[0] < 0:
 *                 f[0] += self.numelems
 *             if not (0 <= f[0] < self.numelems):             # <<<<<<<<<<<<<<
 *                 raise IndexError('Invalid index value: %d, %d' % (
 *                                     f[0], self.numelems))
 */
    __pyx_t_1 = (0 <= (__pyx_v_f[0]));
    if (__pyx_t_1) {
      __pyx_t_1 = ((__pyx_v_f[0]) < __pyx_v_self->numelems);
    }
    __pyx_t_6 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_6) {

      /* "_gappedbuf.pyx":230
 *             if not (0 <= f[0] < self.numelems):
 *                 raise IndexError('Invalid index value: %d, %d' % (
 *                                     f[0], self.numelems))             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_f[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_3 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_1), ((PyObject *)__pyx_t_8)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_7));
      __Pyx_DECREF(((PyObject *)__pyx_t_8)); __pyx_t_8 = 0;
      __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_t_7));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_7));
      __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_Call(__pyx_builtin_IndexError, ((PyObject *)__pyx_t_8), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(((PyObject *)__pyx_t_8)); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L6;
    }
    __pyx_L6:;

    /* "_gappedbuf.pyx":231
 *                 raise IndexError('Invalid index value: %d, %d' % (
 *                                     f[0], self.numelems))
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, object item):
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer._get_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_17__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_17__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_16__getitem__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_item));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":233
 *             return 0
 * 
 *     def __getitem__(self, object item):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t start, stop
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_16__getitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item) {
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "_gappedbuf.pyx":236
 *         cdef Py_ssize_t start, stop
 * 
 *         if self._get_slice(item, &start, &stop):             # <<<<<<<<<<<<<<
 *             return self.get(start, stop)
 *         else:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_get_slice(__pyx_v_self, __pyx_v_item, (&__pyx_v_start), (&__pyx_v_stop)); if (unlikely(PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":237
 * 
 *         if self._get_slice(item, &start, &stop):
 *             return self.get(start, stop)             # <<<<<<<<<<<<<<
 *         else:
 *             return self.buf[start+(0 if start < self.gap else self.gapsize)]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->get(__pyx_v_self, __pyx_v_start, __pyx_v_stop, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":239
 *             return self.get(start, stop)
 *         else:
 *             return self.buf[start+(0 if start < self.gap else self.gapsize)]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, object item, unicode value):
 */
    __Pyx_XDECREF(__pyx_r);
    if (((__pyx_v_start < __pyx_v_self->gap) != 0)) {
      __pyx_t_4 = 0;
    } else {
      __pyx_t_4 = __pyx_v_self->gapsize;
    }
    __pyx_t_3 = PyUnicode_FromOrdinal((__pyx_v_self->buf[(__pyx_v_start + __pyx_t_4)])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_3));
    __pyx_r = ((PyObject *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_19__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_19__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  CYTHON_UNUSED int __pyx_lineno = 0;
  CYTHON_UNUSED const char *__pyx_filename = NULL;
  CYTHON_UNUSED int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 1, "value", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 241; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_18__setitem__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject*)__pyx_v_value));
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":241
 *             return self.buf[start+(0 if start < self.gap else self.gapsize)]
 * 
 *     def __setitem__(self, object item, unicode value):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t start, stop
 * 
 */

static int __pyx_pf_10_gappedbuf_12GappedBuffer_18__setitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_UCS4 __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "_gappedbuf.pyx":244
 *         cdef Py_ssize_t start, stop
 * 
 *         if self._get_slice(item, &start, &stop):             # <<<<<<<<<<<<<<
 *             self.replace(start, stop, value)
 *         else:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_get_slice(__pyx_v_self, __pyx_v_item, (&__pyx_v_start), (&__pyx_v_stop)); if (unlikely(PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":245
 * 
 *         if self._get_slice(item, &start, &stop):
 *             self.replace(start, stop, value)             # <<<<<<<<<<<<<<
 *         else:
 *             if len(value) != 1:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->replace(__pyx_v_self, __pyx_v_start, __pyx_v_stop, __pyx_v_value, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":247
 *             self.replace(start, stop, value)
 *         else:
 *             if len(value) != 1:             # <<<<<<<<<<<<<<
 *                 raise ValueError('Invalid string length: %d' % len(value))
 * 
 */
    if (unlikely(((PyObject *)__pyx_v_value) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_t_4 = __Pyx_PyUnicode_GET_LENGTH(((PyObject *)__pyx_v_value)); if (unlikely(__pyx_t_4 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_2 = ((__pyx_t_4 != 1) != 0);
    if (__pyx_t_2) {

      /* "_gappedbuf.pyx":248
 *         else:
 *             if len(value) != 1:
 *                 raise ValueError('Invalid string length: %d' % len(value))             # <<<<<<<<<<<<<<
 * 
 *             start = start+(0 if start < self.gap else self.gapsize)
 */
      if (unlikely(((PyObject *)__pyx_v_value) == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_t_4 = __Pyx_PyUnicode_GET_LENGTH(((PyObject *)__pyx_v_value)); if (unlikely(__pyx_t_4 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_5), __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_5));
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
      __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L4;
    }
    __pyx_L4:;

    /* "_gappedbuf.pyx":250
 *                 raise ValueError('Invalid string length: %d' % len(value))
 * 
 *             start = start+(0 if start < self.gap else self.gapsize)             # <<<<<<<<<<<<<<
 *             self.buf[start] = <Py_UCS4>value
 * 
 */
    if (((__pyx_v_start < __pyx_v_self->gap) != 0)) {
      __pyx_t_4 = 0;
    } else {
      __pyx_t_4 = __pyx_v_self->gapsize;
    }
    __pyx_v_start = (__pyx_v_start + __pyx_t_4);

    /* "_gappedbuf.pyx":251
 * 
 *             start = start+(0 if start < self.gap else self.gapsize)
 *             self.buf[start] = <Py_UCS4>value             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, object item):
 */
    __pyx_t_6 = __Pyx_PyObject_AsPy_UCS4(((PyObject *)__pyx_v_value)); if (unlikely((__pyx_t_6 == (Py_UCS4)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 251; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    (__pyx_v_self->buf[__pyx_v_start]) = ((Py_UCS4)__pyx_t_6);
  }
  __pyx_L3:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_21__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_21__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_20__delitem__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_item));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":253
 *             self.buf[start] = <Py_UCS4>value
 * 
 *     def __delitem__(self, object item):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t start, stop
 * 
 */

static int __pyx_pf_10_gappedbuf_12GappedBuffer_20__delitem__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_item) {
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "_gappedbuf.pyx":256
 *         cdef Py_ssize_t start, stop
 * 
 *         if self._get_slice(item, &start, &stop):             # <<<<<<<<<<<<<<
 *             self.delete(start, stop)
 *         else:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_get_slice(__pyx_v_self, __pyx_v_item, (&__pyx_v_start), (&__pyx_v_stop)); if (unlikely(PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":257
 * 
 *         if self._get_slice(item, &start, &stop):
 *             self.delete(start, stop)             # <<<<<<<<<<<<<<
 *         else:
 *             self.delete(start, start+1)
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->delete(__pyx_v_self, __pyx_v_start, __pyx_v_stop, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":259
 *             self.delete(start, stop)
 *         else:
 *             self.delete(start, start+1)             # <<<<<<<<<<<<<<
 * 
 *     def findchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->delete(__pyx_v_self, __pyx_v_start, (__pyx_v_start + 1), 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 259; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_23findchr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_23findchr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_c = 0;
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findchr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__c,&__pyx_n_s__begin,&__pyx_n_s__end,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__c)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findchr", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findchr", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findchr") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_c = ((PyObject*)values[0]);
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findchr", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.findchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyUnicode_Type), 1, "c", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 261; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_22findchr(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_c, __pyx_v_begin, __pyx_v_end);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":261
 *             self.delete(start, start+1)
 * 
 *     def findchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         cdef Py_UCS4 *chars
 *         cdef Py_ssize_t size, b, e, p
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_22findchr(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_c, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_UCS4 *__pyx_v_chars;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  long __pyx_v_q;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_UCS4 *__pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("findchr", 0);

  /* "_gappedbuf.pyx":265
 *         cdef Py_ssize_t size, b, e, p
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_1 = (0 <= __pyx_v_begin);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":267
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":269
 *                                 begin, end, self.numelems))
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)             # <<<<<<<<<<<<<<
 *         try:
 *             size = len(c)
 */
  __pyx_t_7 = PyUnicode_AsUCS4Copy(((PyObject *)__pyx_v_c)); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_chars = __pyx_t_7;

  /* "_gappedbuf.pyx":270
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 *         try:             # <<<<<<<<<<<<<<
 *             size = len(c)
 *             # search before gap
 */
  /*try:*/ {

    /* "_gappedbuf.pyx":271
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 *         try:
 *             size = len(c)             # <<<<<<<<<<<<<<
 *             # search before gap
 *             e = min(end, self.gap)
 */
    if (unlikely(((PyObject *)__pyx_v_c) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L5;}
    }
    __pyx_t_8 = __Pyx_PyUnicode_GET_LENGTH(((PyObject *)__pyx_v_c)); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L5;}
    __pyx_v_size = __pyx_t_8;

    /* "_gappedbuf.pyx":273
 *             size = len(c)
 *             # search before gap
 *             e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *             for p from begin <= p < e:
 *                 for q from 0 <= q < size:
 */
    __pyx_t_8 = __pyx_v_self->gap;
    __pyx_t_9 = __pyx_v_end;
    if (((__pyx_t_8 < __pyx_t_9) != 0)) {
      __pyx_t_10 = __pyx_t_8;
    } else {
      __pyx_t_10 = __pyx_t_9;
    }
    __pyx_v_e = __pyx_t_10;

    /* "_gappedbuf.pyx":274
 *             # search before gap
 *             e = min(end, self.gap)
 *             for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 */
    __pyx_t_10 = __pyx_v_e;
    for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_10; __pyx_v_p++) {

      /* "_gappedbuf.pyx":275
 *             e = min(end, self.gap)
 *             for p from begin <= p < e:
 *                 for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                     if self.buf[p] == chars[q]:
 *                         return p
 */
      __pyx_t_8 = __pyx_v_size;
      for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_8; __pyx_v_q++) {

        /* "_gappedbuf.pyx":276
 *             for p from begin <= p < e:
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:             # <<<<<<<<<<<<<<
 *                         return p
 * 
 */
        __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_chars[__pyx_v_q])) != 0);
        if (__pyx_t_2) {

          /* "_gappedbuf.pyx":277
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 *                         return p             # <<<<<<<<<<<<<<
 * 
 *             # search after gap
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L5;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L4;
          goto __pyx_L11;
        }
        __pyx_L11:;
      }
    }

    /* "_gappedbuf.pyx":280
 * 
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *             for p from b <= p < end+self.gapsize:
 *                 for q from 0 <= q < size:
 */
    __pyx_t_10 = __pyx_v_self->gap;
    __pyx_t_8 = __pyx_v_begin;
    if (((__pyx_t_10 > __pyx_t_8) != 0)) {
      __pyx_t_9 = __pyx_t_10;
    } else {
      __pyx_t_9 = __pyx_t_8;
    }
    __pyx_v_b = (__pyx_t_9 + __pyx_v_self->gapsize);

    /* "_gappedbuf.pyx":281
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 */
    __pyx_t_9 = (__pyx_v_end + __pyx_v_self->gapsize);
    for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_9; __pyx_v_p++) {

      /* "_gappedbuf.pyx":282
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from b <= p < end+self.gapsize:
 *                 for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                     if self.buf[p] == chars[q]:
 *                         return p-self.gapsize
 */
      __pyx_t_10 = __pyx_v_size;
      for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

        /* "_gappedbuf.pyx":283
 *             for p from b <= p < end+self.gapsize:
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:             # <<<<<<<<<<<<<<
 *                         return p-self.gapsize
 *         finally:
 */
        __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_chars[__pyx_v_q])) != 0);
        if (__pyx_t_2) {

          /* "_gappedbuf.pyx":284
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 *                         return p-self.gapsize             # <<<<<<<<<<<<<<
 *         finally:
 *             mem.PyMem_Free(chars)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 284; __pyx_clineno = __LINE__; goto __pyx_L5;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L4;
          goto __pyx_L16;
        }
        __pyx_L16:;
      }
    }
  }

  /* "_gappedbuf.pyx":286
 *                         return p-self.gapsize
 *         finally:
 *             mem.PyMem_Free(chars)             # <<<<<<<<<<<<<<
 * 
 *         return -1
 */
  /*finally:*/ {
    int __pyx_why;
    PyObject *__pyx_exc_type, *__pyx_exc_value, *__pyx_exc_tb;
    int __pyx_exc_lineno;
    __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 0; goto __pyx_L6;
    __pyx_L4: __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 3; goto __pyx_L6;
    __pyx_L5: {
      __pyx_why = 4;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_ErrFetch(&__pyx_exc_type, &__pyx_exc_value, &__pyx_exc_tb);
      __pyx_exc_lineno = __pyx_lineno;
      goto __pyx_L6;
    }
    __pyx_L6:;
    PyMem_Free(__pyx_v_chars);
    switch (__pyx_why) {
      case 3: goto __pyx_L0;
      case 4: {
        __Pyx_ErrRestore(__pyx_exc_type, __pyx_exc_value, __pyx_exc_tb);
        __pyx_lineno = __pyx_exc_lineno;
        __pyx_exc_type = 0;
        __pyx_exc_value = 0;
        __pyx_exc_tb = 0;
        goto __pyx_L1_error;
      }
    }
  }

  /* "_gappedbuf.pyx":288
 *             mem.PyMem_Free(chars)
 * 
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     def rfindchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_r = __pyx_int_neg_1;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.findchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_25rfindchr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_25rfindchr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_c = 0;
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rfindchr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__c,&__pyx_n_s__begin,&__pyx_n_s__end,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__c)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rfindchr", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rfindchr", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "rfindchr") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_c = ((PyObject*)values[0]);
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rfindchr", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.rfindchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyUnicode_Type), 1, "c", 1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_24rfindchr(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_c, __pyx_v_begin, __pyx_v_end);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":290
 *         return -1
 * 
 *     def rfindchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 *         cdef Py_UCS4 *chars
 *         cdef Py_ssize_t size, b, e, p
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_24rfindchr(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_c, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  Py_UCS4 *__pyx_v_chars;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  long __pyx_v_q;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_UCS4 *__pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rfindchr", 0);

  /* "_gappedbuf.pyx":294
 *         cdef Py_ssize_t size, b, e, p
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_1 = (0 <= __pyx_v_begin);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":296
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":298
 *                                 begin, end, self.numelems))
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)             # <<<<<<<<<<<<<<
 *         try:
 *             size = len(c)
 */
  __pyx_t_7 = PyUnicode_AsUCS4Copy(((PyObject *)__pyx_v_c)); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_chars = __pyx_t_7;

  /* "_gappedbuf.pyx":299
 * 
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 *         try:             # <<<<<<<<<<<<<<
 *             size = len(c)
 *             # search after gap
 */
  /*try:*/ {

    /* "_gappedbuf.pyx":300
 *         chars = rtdef.PyUnicode_AsUCS4Copy(c)
 *         try:
 *             size = len(c)             # <<<<<<<<<<<<<<
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize
 */
    if (unlikely(((PyObject *)__pyx_v_c) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 300; __pyx_clineno = __LINE__; goto __pyx_L5;}
    }
    __pyx_t_8 = __Pyx_PyUnicode_GET_LENGTH(((PyObject *)__pyx_v_c)); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 300; __pyx_clineno = __LINE__; goto __pyx_L5;}
    __pyx_v_size = __pyx_t_8;

    /* "_gappedbuf.pyx":302
 *             size = len(c)
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *             for p from (end+self.gapsize) > p >= b:
 *                 for q from 0 <= q < size:
 */
    __pyx_t_8 = __pyx_v_self->gap;
    __pyx_t_9 = __pyx_v_begin;
    if (((__pyx_t_8 > __pyx_t_9) != 0)) {
      __pyx_t_10 = __pyx_t_8;
    } else {
      __pyx_t_10 = __pyx_t_9;
    }
    __pyx_v_b = (__pyx_t_10 + __pyx_v_self->gapsize);

    /* "_gappedbuf.pyx":303
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from (end+self.gapsize) > p >= b:             # <<<<<<<<<<<<<<
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 */
    __pyx_t_10 = __pyx_v_b;
    for (__pyx_v_p = (__pyx_v_end + __pyx_v_self->gapsize)-1; __pyx_v_p >= __pyx_t_10; __pyx_v_p--) {

      /* "_gappedbuf.pyx":304
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from (end+self.gapsize) > p >= b:
 *                 for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                     if self.buf[p] == chars[q]:
 *                         return p-self.gapsize
 */
      __pyx_t_8 = __pyx_v_size;
      for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_8; __pyx_v_q++) {

        /* "_gappedbuf.pyx":305
 *             for p from (end+self.gapsize) > p >= b:
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:             # <<<<<<<<<<<<<<
 *                         return p-self.gapsize
 * 
 */
        __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_chars[__pyx_v_q])) != 0);
        if (__pyx_t_2) {

          /* "_gappedbuf.pyx":306
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 *                         return p-self.gapsize             # <<<<<<<<<<<<<<
 * 
 *             # search before gap
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 306; __pyx_clineno = __LINE__; goto __pyx_L5;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L4;
          goto __pyx_L11;
        }
        __pyx_L11:;
      }
    }

    /* "_gappedbuf.pyx":309
 * 
 *             # search before gap
 *             e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *             for p from e > p >= begin:
 *                 for q from 0 <= q < size:
 */
    __pyx_t_10 = __pyx_v_self->gap;
    __pyx_t_8 = __pyx_v_end;
    if (((__pyx_t_10 < __pyx_t_8) != 0)) {
      __pyx_t_9 = __pyx_t_10;
    } else {
      __pyx_t_9 = __pyx_t_8;
    }
    __pyx_v_e = __pyx_t_9;

    /* "_gappedbuf.pyx":310
 *             # search before gap
 *             e = min(end, self.gap)
 *             for p from e > p >= begin:             # <<<<<<<<<<<<<<
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 */
    __pyx_t_9 = __pyx_v_begin;
    for (__pyx_v_p = __pyx_v_e-1; __pyx_v_p >= __pyx_t_9; __pyx_v_p--) {

      /* "_gappedbuf.pyx":311
 *             e = min(end, self.gap)
 *             for p from e > p >= begin:
 *                 for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                     if self.buf[p] == chars[q]:
 *                         return p
 */
      __pyx_t_10 = __pyx_v_size;
      for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

        /* "_gappedbuf.pyx":312
 *             for p from e > p >= begin:
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:             # <<<<<<<<<<<<<<
 *                         return p
 *         finally:
 */
        __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_chars[__pyx_v_q])) != 0);
        if (__pyx_t_2) {

          /* "_gappedbuf.pyx":313
 *                 for q from 0 <= q < size:
 *                     if self.buf[p] == chars[q]:
 *                         return p             # <<<<<<<<<<<<<<
 *         finally:
 *             mem.PyMem_Free(chars)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L5;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L4;
          goto __pyx_L16;
        }
        __pyx_L16:;
      }
    }
  }

  /* "_gappedbuf.pyx":315
 *                         return p
 *         finally:
 *             mem.PyMem_Free(chars)             # <<<<<<<<<<<<<<
 * 
 *         return -1
 */
  /*finally:*/ {
    int __pyx_why;
    PyObject *__pyx_exc_type, *__pyx_exc_value, *__pyx_exc_tb;
    int __pyx_exc_lineno;
    __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 0; goto __pyx_L6;
    __pyx_L4: __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 3; goto __pyx_L6;
    __pyx_L5: {
      __pyx_why = 4;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_ErrFetch(&__pyx_exc_type, &__pyx_exc_value, &__pyx_exc_tb);
      __pyx_exc_lineno = __pyx_lineno;
      goto __pyx_L6;
    }
    __pyx_L6:;
    PyMem_Free(__pyx_v_chars);
    switch (__pyx_why) {
      case 3: goto __pyx_L0;
      case 4: {
        __Pyx_ErrRestore(__pyx_exc_type, __pyx_exc_value, __pyx_exc_tb);
        __pyx_lineno = __pyx_exc_lineno;
        __pyx_exc_type = 0;
        __pyx_exc_value = 0;
        __pyx_exc_tb = 0;
        goto __pyx_L1_error;
      }
    }
  }

  /* "_gappedbuf.pyx":317
 *             mem.PyMem_Free(chars)
 * 
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     cpdef getint(self, Py_ssize_t pos):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_r = __pyx_int_neg_1;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.rfindchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":319
 *         return -1
 * 
 *     cpdef getint(self, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *         if not (0 <= pos < self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_27getint(PyObject *__pyx_v_self, PyObject *__pyx_arg_pos); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_getint(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_pos, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getint", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__getint); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_27getint)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_pos); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":320
 * 
 *     cpdef getint(self, Py_ssize_t pos):
 *         if not (0 <= pos < self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 pos, self.numelems))
 */
  __pyx_t_4 = (0 <= __pyx_v_pos);
  if (__pyx_t_4) {
    __pyx_t_4 = (__pyx_v_pos < __pyx_v_self->numelems);
  }
  __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_5) {

    /* "_gappedbuf.pyx":322
 *         if not (0 <= pos < self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 pos, self.numelems))             # <<<<<<<<<<<<<<
 *         if pos < self.gap:
 *             return <unsigned long>self.buf[pos]
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_pos); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 322; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 322; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 322; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_1), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_2));
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_2));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
    __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":323
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 pos, self.numelems))
 *         if pos < self.gap:             # <<<<<<<<<<<<<<
 *             return <unsigned long>self.buf[pos]
 *         else:
 */
  __pyx_t_5 = ((__pyx_v_pos < __pyx_v_self->gap) != 0);
  if (__pyx_t_5) {

    /* "_gappedbuf.pyx":324
 *                                 pos, self.numelems))
 *         if pos < self.gap:
 *             return <unsigned long>self.buf[pos]             # <<<<<<<<<<<<<<
 *         else:
 *             return <unsigned long>self.buf[self.gapsize+pos]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyLong_FromUnsignedLong(((unsigned long)(__pyx_v_self->buf[__pyx_v_pos]))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 324; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "_gappedbuf.pyx":326
 *             return <unsigned long>self.buf[pos]
 *         else:
 *             return <unsigned long>self.buf[self.gapsize+pos]             # <<<<<<<<<<<<<<
 * 
 *     cpdef getints(self, Py_ssize_t begin, Py_ssize_t end):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyLong_FromUnsignedLong(((unsigned long)(__pyx_v_self->buf[(__pyx_v_self->gapsize + __pyx_v_pos)]))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 326; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __pyx_L4:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_27getint(PyObject *__pyx_v_self, PyObject *__pyx_arg_pos); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_27getint(PyObject *__pyx_v_self, PyObject *__pyx_arg_pos) {
  Py_ssize_t __pyx_v_pos;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getint (wrapper)", 0);
  assert(__pyx_arg_pos); {
    __pyx_v_pos = __Pyx_PyIndex_AsSsize_t(__pyx_arg_pos); if (unlikely((__pyx_v_pos == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_26getint(((struct GappedBufferObj *)__pyx_v_self), ((Py_ssize_t)__pyx_v_pos));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":319
 *         return -1
 * 
 *     cpdef getint(self, Py_ssize_t pos):             # <<<<<<<<<<<<<<
 *         if not (0 <= pos < self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_26getint(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_pos) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getint", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->getint(__pyx_v_self, __pyx_v_pos, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":328
 *             return <unsigned long>self.buf[self.gapsize+pos]
 * 
 *     cpdef getints(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t b, e, p, n
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_29getints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_getints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_ret = 0;
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getints", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__getints); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_29getints)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":334
 *         cdef object v
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_5 = (0 <= __pyx_v_begin);
  if (__pyx_t_5) {
    __pyx_t_5 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_5) {
      __pyx_t_5 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
  if (__pyx_t_6) {

    /* "_gappedbuf.pyx":336
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         ret = PyList_New(end-begin)
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_4));
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_4));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":338
 *                                 begin, end, self.numelems))
 * 
 *         ret = PyList_New(end-begin)             # <<<<<<<<<<<<<<
 *         n = 0
 *         # read before gap
 */
  __pyx_t_4 = PyList_New((__pyx_v_end - __pyx_v_begin)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 338; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected list, got %.200s", Py_TYPE(__pyx_t_4)->tp_name), 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 338; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_ret = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "_gappedbuf.pyx":339
 * 
 *         ret = PyList_New(end-begin)
 *         n = 0             # <<<<<<<<<<<<<<
 *         # read before gap
 *         e = min(end, self.gap)
 */
  __pyx_v_n = 0;

  /* "_gappedbuf.pyx":341
 *         n = 0
 *         # read before gap
 *         e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *         for p from begin <= p < e:
 *             v = <unsigned long>self.buf[p]
 */
  __pyx_t_7 = __pyx_v_self->gap;
  __pyx_t_8 = __pyx_v_end;
  if (((__pyx_t_7 < __pyx_t_8) != 0)) {
    __pyx_t_9 = __pyx_t_7;
  } else {
    __pyx_t_9 = __pyx_t_8;
  }
  __pyx_v_e = __pyx_t_9;

  /* "_gappedbuf.pyx":342
 *         # read before gap
 *         e = min(end, self.gap)
 *         for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)
 */
  __pyx_t_9 = __pyx_v_e;
  for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_9; __pyx_v_p++) {

    /* "_gappedbuf.pyx":343
 *         e = min(end, self.gap)
 *         for p from begin <= p < e:
 *             v = <unsigned long>self.buf[p]             # <<<<<<<<<<<<<<
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)
 */
    __pyx_t_4 = PyLong_FromUnsignedLong(((unsigned long)(__pyx_v_self->buf[__pyx_v_p]))); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 343; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "_gappedbuf.pyx":344
 *         for p from begin <= p < e:
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)             # <<<<<<<<<<<<<<
 *             ref.Py_INCREF(v)
 *             n += 1
 */
    PyList_SET_ITEM(((PyObject *)__pyx_v_ret), __pyx_v_n, __pyx_v_v);

    /* "_gappedbuf.pyx":345
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)             # <<<<<<<<<<<<<<
 *             n += 1
 * 
 */
    Py_INCREF(__pyx_v_v);

    /* "_gappedbuf.pyx":346
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)
 *             n += 1             # <<<<<<<<<<<<<<
 * 
 *         # read after gap
 */
    __pyx_v_n = (__pyx_v_n + 1);
  }

  /* "_gappedbuf.pyx":349
 * 
 *         # read after gap
 *         b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *         for p from b <= p < end+self.gapsize:
 *             v = <unsigned long>self.buf[p]
 */
  __pyx_t_9 = __pyx_v_self->gap;
  __pyx_t_7 = __pyx_v_begin;
  if (((__pyx_t_9 > __pyx_t_7) != 0)) {
    __pyx_t_8 = __pyx_t_9;
  } else {
    __pyx_t_8 = __pyx_t_7;
  }
  __pyx_v_b = (__pyx_t_8 + __pyx_v_self->gapsize);

  /* "_gappedbuf.pyx":350
 *         # read after gap
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)
 */
  __pyx_t_8 = (__pyx_v_end + __pyx_v_self->gapsize);
  for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_8; __pyx_v_p++) {

    /* "_gappedbuf.pyx":351
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:
 *             v = <unsigned long>self.buf[p]             # <<<<<<<<<<<<<<
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)
 */
    __pyx_t_4 = PyLong_FromUnsignedLong(((unsigned long)(__pyx_v_self->buf[__pyx_v_p]))); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 351; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "_gappedbuf.pyx":352
 *         for p from b <= p < end+self.gapsize:
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)             # <<<<<<<<<<<<<<
 *             ref.Py_INCREF(v)
 *             n += 1
 */
    PyList_SET_ITEM(((PyObject *)__pyx_v_ret), __pyx_v_n, __pyx_v_v);

    /* "_gappedbuf.pyx":353
 *             v = <unsigned long>self.buf[p]
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)             # <<<<<<<<<<<<<<
 *             n += 1
 * 
 */
    Py_INCREF(__pyx_v_v);

    /* "_gappedbuf.pyx":354
 *             PyList_SET_ITEM(ret, n, v)
 *             ref.Py_INCREF(v)
 *             n += 1             # <<<<<<<<<<<<<<
 * 
 *         return ret
 */
    __pyx_v_n = (__pyx_v_n + 1);
  }

  /* "_gappedbuf.pyx":356
 *             n += 1
 * 
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     cdef _insertints(self, Py_ssize_t index, object s):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_ret));
  __pyx_r = ((PyObject *)__pyx_v_ret);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_29getints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_29getints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getints (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getints", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getints") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getints", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_28getints(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":328
 *             return <unsigned long>self.buf[self.gapsize+pos]
 * 
 *     cpdef getints(self, Py_ssize_t begin, Py_ssize_t end):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t b, e, p, n
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_28getints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getints", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->getints(__pyx_v_self, __pyx_v_begin, __pyx_v_end, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.getints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":358
 *         return ret
 * 
 *     cdef _insertints(self, Py_ssize_t index, object s):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t size = len(s)
 */

static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer__insertints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s) {
  Py_ssize_t __pyx_v_size;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  unsigned long __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_insertints", 0);

  /* "_gappedbuf.pyx":360
 *     cdef _insertints(self, Py_ssize_t index, object s):
 * 
 *         cdef Py_ssize_t size = len(s)             # <<<<<<<<<<<<<<
 *         if not (0 <= index <= self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_s); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_size = __pyx_t_1;

  /* "_gappedbuf.pyx":361
 * 
 *         cdef Py_ssize_t size = len(s)
 *         if not (0 <= index <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 index, self.numelems))
 */
  __pyx_t_2 = (0 <= __pyx_v_index);
  if (__pyx_t_2) {
    __pyx_t_2 = (__pyx_v_index <= __pyx_v_self->numelems);
  }
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "_gappedbuf.pyx":363
 *         if not (0 <= index <= self.numelems):
 *             raise ValueError('Invalid index value: %d, %d' % (
 *                                 index, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         if size:
 */
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_index); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_1), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":365
 *                                 index, self.numelems))
 * 
 *         if size:             # <<<<<<<<<<<<<<
 *             self._move_gap(index)
 *             if self.gapsize < size:
 */
  __pyx_t_3 = (__pyx_v_size != 0);
  if (__pyx_t_3) {

    /* "_gappedbuf.pyx":366
 * 
 *         if size:
 *             self._move_gap(index)             # <<<<<<<<<<<<<<
 *             if self.gapsize < size:
 *                 self._expand_gap(max(1024, size*2))
 */
    ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_move_gap(__pyx_v_self, __pyx_v_index);

    /* "_gappedbuf.pyx":367
 *         if size:
 *             self._move_gap(index)
 *             if self.gapsize < size:             # <<<<<<<<<<<<<<
 *                 self._expand_gap(max(1024, size*2))
 * 
 */
    __pyx_t_3 = ((__pyx_v_self->gapsize < __pyx_v_size) != 0);
    if (__pyx_t_3) {

      /* "_gappedbuf.pyx":368
 *             self._move_gap(index)
 *             if self.gapsize < size:
 *                 self._expand_gap(max(1024, size*2))             # <<<<<<<<<<<<<<
 * 
 *             for value in s:
 */
      __pyx_t_1 = (__pyx_v_size * 2);
      __pyx_t_7 = 1024;
      if (((__pyx_t_1 > __pyx_t_7) != 0)) {
        __pyx_t_8 = __pyx_t_1;
      } else {
        __pyx_t_8 = __pyx_t_7;
      }
      ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_expand_gap(__pyx_v_self, __pyx_t_8);
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "_gappedbuf.pyx":370
 *                 self._expand_gap(max(1024, size*2))
 * 
 *             for value in s:             # <<<<<<<<<<<<<<
 *                 self.buf[self.gap] = <unsigned long>value
 *                 self.gap += 1
 */
    if (PyList_CheckExact(__pyx_v_s) || PyTuple_CheckExact(__pyx_v_s)) {
      __pyx_t_5 = __pyx_v_s; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_s); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext;
    }
    for (;;) {
      if (!__pyx_t_9 && PyList_CheckExact(__pyx_t_5)) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_9 && PyTuple_CheckExact(__pyx_t_5)) {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_6 = __pyx_t_9(__pyx_t_5);
        if (unlikely(!__pyx_t_6)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "_gappedbuf.pyx":371
 * 
 *             for value in s:
 *                 self.buf[self.gap] = <unsigned long>value             # <<<<<<<<<<<<<<
 *                 self.gap += 1
 * 
 */
      __pyx_t_10 = __Pyx_PyInt_AsUnsignedLong(__pyx_v_value); if (unlikely((__pyx_t_10 == (unsigned long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 371; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      (__pyx_v_self->buf[__pyx_v_self->gap]) = ((unsigned long)__pyx_t_10);

      /* "_gappedbuf.pyx":372
 *             for value in s:
 *                 self.buf[self.gap] = <unsigned long>value
 *                 self.gap += 1             # <<<<<<<<<<<<<<
 * 
 *             self.numelems += size
 */
      __pyx_v_self->gap = (__pyx_v_self->gap + 1);
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_gappedbuf.pyx":374
 *                 self.gap += 1
 * 
 *             self.numelems += size             # <<<<<<<<<<<<<<
 *             self.gapsize -= size
 * 
 */
    __pyx_v_self->numelems = (__pyx_v_self->numelems + __pyx_v_size);

    /* "_gappedbuf.pyx":375
 * 
 *             self.numelems += size
 *             self.gapsize -= size             # <<<<<<<<<<<<<<
 * 
 *     cpdef insertints(self, Py_ssize_t index, object s):
 */
    __pyx_v_self->gapsize = (__pyx_v_self->gapsize - __pyx_v_size);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer._insertints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":377
 *             self.gapsize -= size
 * 
 *     cpdef insertints(self, Py_ssize_t index, object s):             # <<<<<<<<<<<<<<
 * 
 *         self._insertints(index, s)
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_31insertints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_insertints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insertints", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__insertints); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_31insertints)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_index); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":379
 *     cpdef insertints(self, Py_ssize_t index, object s):
 * 
 *         self._insertints(index, s)             # <<<<<<<<<<<<<<
 * 
 *     cpdef appendints(self, object s):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_insertints(__pyx_v_self, __pyx_v_index, __pyx_v_s); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 379; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insertints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_31insertints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_31insertints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_v_s = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insertints (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__index,&__pyx_n_s__s,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insertints", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insertints") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_index = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_index == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_s = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insertints", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insertints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_30insertints(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_index, __pyx_v_s);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":377
 *             self.gapsize -= size
 * 
 *     cpdef insertints(self, Py_ssize_t index, object s):             # <<<<<<<<<<<<<<
 * 
 *         self._insertints(index, s)
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_30insertints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_index, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insertints", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->insertints(__pyx_v_self, __pyx_v_index, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 377; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.insertints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":381
 *         self._insertints(index, s)
 * 
 *     cpdef appendints(self, object s):             # <<<<<<<<<<<<<<
 *         self.insertints(self.numelems, s)
 * 
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_33appendints(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_appendints(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("appendints", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__appendints); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 381; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_33appendints)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 381; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 381; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":382
 * 
 *     cpdef appendints(self, object s):
 *         self.insertints(self.numelems, s)             # <<<<<<<<<<<<<<
 * 
 *     cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->insertints(__pyx_v_self, __pyx_v_self->numelems, __pyx_v_s, 0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 382; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.appendints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_33appendints(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_33appendints(PyObject *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("appendints (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_32appendints(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_s));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":381
 *         self._insertints(index, s)
 * 
 *     cpdef appendints(self, object s):             # <<<<<<<<<<<<<<
 *         self.insertints(self.numelems, s)
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_32appendints(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("appendints", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->appendints(__pyx_v_self, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 381; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.appendints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":384
 *         self.insertints(self.numelems, s)
 * 
 *     cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 *         self._insertints(begin, s)
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_35replaceints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_replaceints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replaceints", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__replaceints); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_35replaceints)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":385
 * 
 *     cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):
 *         self._delete(begin, end)             # <<<<<<<<<<<<<<
 *         self._insertints(begin, s)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_delete(__pyx_v_self, __pyx_v_begin, __pyx_v_end); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_gappedbuf.pyx":386
 *     cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):
 *         self._delete(begin, end)
 *         self._insertints(begin, s)             # <<<<<<<<<<<<<<
 * 
 *     cpdef setints(self, Py_ssize_t begin, Py_ssize_t end, unsigned long v):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->_insertints(__pyx_v_self, __pyx_v_begin, __pyx_v_s); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replaceints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_35replaceints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_35replaceints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  PyObject *__pyx_v_s = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replaceints (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,&__pyx_n_s__s,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("replaceints", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("replaceints", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "replaceints") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_s = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("replaceints", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replaceints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_34replaceints(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end, __pyx_v_s);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":384
 *         self.insertints(self.numelems, s)
 * 
 *     cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):             # <<<<<<<<<<<<<<
 *         self._delete(begin, end)
 *         self._insertints(begin, s)
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_34replaceints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replaceints", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->replaceints(__pyx_v_self, __pyx_v_begin, __pyx_v_end, __pyx_v_s, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 384; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.replaceints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":388
 *         self._insertints(begin, s)
 * 
 *     cpdef setints(self, Py_ssize_t begin, Py_ssize_t end, unsigned long v):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t b, e, p
 * 
 */

static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_37setints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10_gappedbuf_12GappedBuffer_setints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, unsigned long __pyx_v_v, int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  CYTHON_UNUSED long __pyx_v_n;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setints", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s__setints); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_37setints)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyLong_FromUnsignedLong(__pyx_v_v); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "_gappedbuf.pyx":391
 *         cdef Py_ssize_t b, e, p
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_6 = (0 <= __pyx_v_begin);
  if (__pyx_t_6) {
    __pyx_t_6 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_6) {
      __pyx_t_6 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_7 = ((!(__pyx_t_6 != 0)) != 0);
  if (__pyx_t_7) {

    /* "_gappedbuf.pyx":393
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         n = 0
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":395
 *                                 begin, end, self.numelems))
 * 
 *         n = 0             # <<<<<<<<<<<<<<
 *         # replace before gap
 *         e = min(end, self.gap)
 */
  __pyx_v_n = 0;

  /* "_gappedbuf.pyx":397
 *         n = 0
 *         # replace before gap
 *         e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *         for p from begin <= p < e:
 *             self.buf[p] = v
 */
  __pyx_t_8 = __pyx_v_self->gap;
  __pyx_t_9 = __pyx_v_end;
  if (((__pyx_t_8 < __pyx_t_9) != 0)) {
    __pyx_t_10 = __pyx_t_8;
  } else {
    __pyx_t_10 = __pyx_t_9;
  }
  __pyx_v_e = __pyx_t_10;

  /* "_gappedbuf.pyx":398
 *         # replace before gap
 *         e = min(end, self.gap)
 *         for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *             self.buf[p] = v
 * 
 */
  __pyx_t_10 = __pyx_v_e;
  for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_10; __pyx_v_p++) {

    /* "_gappedbuf.pyx":399
 *         e = min(end, self.gap)
 *         for p from begin <= p < e:
 *             self.buf[p] = v             # <<<<<<<<<<<<<<
 * 
 *         # replace after gap
 */
    (__pyx_v_self->buf[__pyx_v_p]) = __pyx_v_v;
  }

  /* "_gappedbuf.pyx":402
 * 
 *         # replace after gap
 *         b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *         for p from b <= p < end+self.gapsize:
 *             self.buf[p] = v
 */
  __pyx_t_10 = __pyx_v_self->gap;
  __pyx_t_8 = __pyx_v_begin;
  if (((__pyx_t_10 > __pyx_t_8) != 0)) {
    __pyx_t_9 = __pyx_t_10;
  } else {
    __pyx_t_9 = __pyx_t_8;
  }
  __pyx_v_b = (__pyx_t_9 + __pyx_v_self->gapsize);

  /* "_gappedbuf.pyx":403
 *         # replace after gap
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *             self.buf[p] = v
 * 
 */
  __pyx_t_9 = (__pyx_v_end + __pyx_v_self->gapsize);
  for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_9; __pyx_v_p++) {

    /* "_gappedbuf.pyx":404
 *         b = max(begin, self.gap) + self.gapsize
 *         for p from b <= p < end+self.gapsize:
 *             self.buf[p] = v             # <<<<<<<<<<<<<<
 * 
 *     def findint(self, object nums, Py_ssize_t begin, Py_ssize_t end,
 */
    (__pyx_v_self->buf[__pyx_v_p]) = __pyx_v_v;
  }

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.setints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_37setints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_37setints(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  unsigned long __pyx_v_v;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setints (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__begin,&__pyx_n_s__end,&__pyx_n_s__v,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setints", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setints", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setints") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[0]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_v = __Pyx_PyInt_AsUnsignedLong(values[2]); if (unlikely((__pyx_v_v == (unsigned long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setints", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.setints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_36setints(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_begin, __pyx_v_end, __pyx_v_v);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":388
 *         self._insertints(begin, s)
 * 
 *     cpdef setints(self, Py_ssize_t begin, Py_ssize_t end, unsigned long v):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t b, e, p
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_36setints(struct GappedBufferObj *__pyx_v_self, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, unsigned long __pyx_v_v) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setints", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer *)__pyx_v_self->__pyx_vtab)->setints(__pyx_v_self, __pyx_v_begin, __pyx_v_end, __pyx_v_v, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.setints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_39findint(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_39findint(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nums = 0;
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_v_comp_ne;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findint (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__nums,&__pyx_n_s__begin,&__pyx_n_s__end,&__pyx_n_s__comp_ne,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__nums)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findint", 1, 4, 4, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findint", 1, 4, 4, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__comp_ne)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findint", 1, 4, 4, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findint") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_nums = values[0];
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_comp_ne = __Pyx_PyInt_AsInt(values[3]); if (unlikely((__pyx_v_comp_ne == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findint", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.findint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_38findint(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_nums, __pyx_v_begin, __pyx_v_end, __pyx_v_comp_ne);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":406
 *             self.buf[p] = v
 * 
 *     def findint(self, object nums, Py_ssize_t begin, Py_ssize_t end,             # <<<<<<<<<<<<<<
 *                 int comp_ne):
 *         cdef unsigned long *longs
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_38findint(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_nums, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_v_comp_ne) {
  unsigned long *__pyx_v_longs;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  long __pyx_v_i;
  long __pyx_v_q;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  unsigned long __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("findint", 0);

  /* "_gappedbuf.pyx":411
 *         cdef Py_ssize_t size, b, e, p
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_1 = (0 <= __pyx_v_begin);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":413
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         size = len(nums)
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":415
 *                                 begin, end, self.numelems))
 * 
 *         size = len(nums)             # <<<<<<<<<<<<<<
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_nums); if (unlikely(__pyx_t_7 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 415; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_size = __pyx_t_7;

  /* "_gappedbuf.pyx":417
 *         size = len(nums)
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))             # <<<<<<<<<<<<<<
 *         if not longs:
 *             raise MemoryError()
 */
  __pyx_v_longs = ((unsigned long *)PyMem_Malloc((__pyx_v_size * (sizeof(unsigned long)))));

  /* "_gappedbuf.pyx":418
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))
 *         if not longs:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_longs != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":419
 *                                     size * sizeof(unsigned long))
 *         if not longs:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 419; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":421
 *             raise MemoryError()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < size:
 *                 longs[i] = nums[i]
 */
  /*try:*/ {

    /* "_gappedbuf.pyx":422
 * 
 *         try:
 *             for i from 0 <= i < size:             # <<<<<<<<<<<<<<
 *                 longs[i] = nums[i]
 * 
 */
    __pyx_t_7 = __pyx_v_size;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_7; __pyx_v_i++) {

      /* "_gappedbuf.pyx":423
 *         try:
 *             for i from 0 <= i < size:
 *                 longs[i] = nums[i]             # <<<<<<<<<<<<<<
 * 
 *             # search before gap
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_nums, __pyx_v_i, sizeof(long), PyInt_FromLong, 0, 1, 1); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 423; __pyx_clineno = __LINE__; goto __pyx_L6;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyInt_AsUnsignedLong(__pyx_t_5); if (unlikely((__pyx_t_8 == (unsigned long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 423; __pyx_clineno = __LINE__; goto __pyx_L6;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      (__pyx_v_longs[__pyx_v_i]) = __pyx_t_8;
    }

    /* "_gappedbuf.pyx":426
 * 
 *             # search before gap
 *             e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *             for p from begin <= p < e:
 *                 if not comp_ne:
 */
    __pyx_t_7 = __pyx_v_self->gap;
    __pyx_t_9 = __pyx_v_end;
    if (((__pyx_t_7 < __pyx_t_9) != 0)) {
      __pyx_t_10 = __pyx_t_7;
    } else {
      __pyx_t_10 = __pyx_t_9;
    }
    __pyx_v_e = __pyx_t_10;

    /* "_gappedbuf.pyx":427
 *             # search before gap
 *             e = min(end, self.gap)
 *             for p from begin <= p < e:             # <<<<<<<<<<<<<<
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 */
    __pyx_t_10 = __pyx_v_e;
    for (__pyx_v_p = __pyx_v_begin; __pyx_v_p < __pyx_t_10; __pyx_v_p++) {

      /* "_gappedbuf.pyx":428
 *             e = min(end, self.gap)
 *             for p from begin <= p < e:
 *                 if not comp_ne:             # <<<<<<<<<<<<<<
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 */
      __pyx_t_2 = ((!(__pyx_v_comp_ne != 0)) != 0);
      if (__pyx_t_2) {

        /* "_gappedbuf.pyx":429
 *             for p from begin <= p < e:
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             return p
 */
        __pyx_t_7 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_7; __pyx_v_q++) {

          /* "_gappedbuf.pyx":430
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             return p
 *                 else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":431
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             return p             # <<<<<<<<<<<<<<
 *                 else:
 *                     for q from 0 <= q < size:
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 431; __pyx_clineno = __LINE__; goto __pyx_L6;}
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_r = __pyx_t_5;
            __pyx_t_5 = 0;
            goto __pyx_L5;
            goto __pyx_L15;
          }
          __pyx_L15:;
        }
        goto __pyx_L12;
      }
      /*else*/ {

        /* "_gappedbuf.pyx":433
 *                             return p
 *                 else:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             break
 */
        __pyx_t_7 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_7; __pyx_v_q++) {

          /* "_gappedbuf.pyx":434
 *                 else:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             break
 *                     else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":435
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             break             # <<<<<<<<<<<<<<
 *                     else:
 *                         return p
 */
            goto __pyx_L17_break;
            goto __pyx_L18;
          }
          __pyx_L18:;
        }
        /*else*/ {

          /* "_gappedbuf.pyx":437
 *                             break
 *                     else:
 *                         return p             # <<<<<<<<<<<<<<
 * 
 *             # search after gap
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 437; __pyx_clineno = __LINE__; goto __pyx_L6;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L5;
        }
        __pyx_L17_break:;
      }
      __pyx_L12:;
    }

    /* "_gappedbuf.pyx":440
 * 
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *             for p from b <= p < end+self.gapsize:
 *                 if not comp_ne:
 */
    __pyx_t_10 = __pyx_v_self->gap;
    __pyx_t_7 = __pyx_v_begin;
    if (((__pyx_t_10 > __pyx_t_7) != 0)) {
      __pyx_t_9 = __pyx_t_10;
    } else {
      __pyx_t_9 = __pyx_t_7;
    }
    __pyx_v_b = (__pyx_t_9 + __pyx_v_self->gapsize);

    /* "_gappedbuf.pyx":441
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from b <= p < end+self.gapsize:             # <<<<<<<<<<<<<<
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 */
    __pyx_t_9 = (__pyx_v_end + __pyx_v_self->gapsize);
    for (__pyx_v_p = __pyx_v_b; __pyx_v_p < __pyx_t_9; __pyx_v_p++) {

      /* "_gappedbuf.pyx":442
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from b <= p < end+self.gapsize:
 *                 if not comp_ne:             # <<<<<<<<<<<<<<
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 */
      __pyx_t_2 = ((!(__pyx_v_comp_ne != 0)) != 0);
      if (__pyx_t_2) {

        /* "_gappedbuf.pyx":443
 *             for p from b <= p < end+self.gapsize:
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             return p-self.gapsize
 */
        __pyx_t_10 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

          /* "_gappedbuf.pyx":444
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             return p-self.gapsize
 *                 else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":445
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             return p-self.gapsize             # <<<<<<<<<<<<<<
 *                 else:
 *                     for q from 0 <= q < size:
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L6;}
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_r = __pyx_t_5;
            __pyx_t_5 = 0;
            goto __pyx_L5;
            goto __pyx_L24;
          }
          __pyx_L24:;
        }
        goto __pyx_L21;
      }
      /*else*/ {

        /* "_gappedbuf.pyx":447
 *                             return p-self.gapsize
 *                 else:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             break
 */
        __pyx_t_10 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

          /* "_gappedbuf.pyx":448
 *                 else:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             break
 *                     else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":449
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             break             # <<<<<<<<<<<<<<
 *                     else:
 *                         return p-self.gapsize
 */
            goto __pyx_L26_break;
            goto __pyx_L27;
          }
          __pyx_L27:;
        }
        /*else*/ {

          /* "_gappedbuf.pyx":451
 *                             break
 *                     else:
 *                         return p-self.gapsize             # <<<<<<<<<<<<<<
 *         finally:
 *             mem.PyMem_Free(longs)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 451; __pyx_clineno = __LINE__; goto __pyx_L6;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L5;
        }
        __pyx_L26_break:;
      }
      __pyx_L21:;
    }
  }

  /* "_gappedbuf.pyx":453
 *                         return p-self.gapsize
 *         finally:
 *             mem.PyMem_Free(longs)             # <<<<<<<<<<<<<<
 * 
 *         return -1
 */
  /*finally:*/ {
    int __pyx_why;
    PyObject *__pyx_exc_type, *__pyx_exc_value, *__pyx_exc_tb;
    int __pyx_exc_lineno;
    __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 0; goto __pyx_L7;
    __pyx_L5: __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 3; goto __pyx_L7;
    __pyx_L6: {
      __pyx_why = 4;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_ErrFetch(&__pyx_exc_type, &__pyx_exc_value, &__pyx_exc_tb);
      __pyx_exc_lineno = __pyx_lineno;
      goto __pyx_L7;
    }
    __pyx_L7:;
    PyMem_Free(__pyx_v_longs);
    switch (__pyx_why) {
      case 3: goto __pyx_L0;
      case 4: {
        __Pyx_ErrRestore(__pyx_exc_type, __pyx_exc_value, __pyx_exc_tb);
        __pyx_lineno = __pyx_exc_lineno;
        __pyx_exc_type = 0;
        __pyx_exc_value = 0;
        __pyx_exc_tb = 0;
        goto __pyx_L1_error;
      }
    }
  }

  /* "_gappedbuf.pyx":455
 *             mem.PyMem_Free(longs)
 * 
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     def rfindint(self, object nums, Py_ssize_t begin, Py_ssize_t end,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_r = __pyx_int_neg_1;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.findint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_41rfindint(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_41rfindint(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nums = 0;
  Py_ssize_t __pyx_v_begin;
  Py_ssize_t __pyx_v_end;
  int __pyx_v_comp_ne;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rfindint (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__nums,&__pyx_n_s__begin,&__pyx_n_s__end,&__pyx_n_s__comp_ne,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__nums)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__begin)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rfindint", 1, 4, 4, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rfindint", 1, 4, 4, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__comp_ne)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rfindint", 1, 4, 4, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "rfindint") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_nums = values[0];
    __pyx_v_begin = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_begin == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_end = __Pyx_PyIndex_AsSsize_t(values[2]); if (unlikely((__pyx_v_end == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_comp_ne = __Pyx_PyInt_AsInt(values[3]); if (unlikely((__pyx_v_comp_ne == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 458; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rfindint", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.rfindint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_40rfindint(((struct GappedBufferObj *)__pyx_v_self), __pyx_v_nums, __pyx_v_begin, __pyx_v_end, __pyx_v_comp_ne);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pyx":457
 *         return -1
 * 
 *     def rfindint(self, object nums, Py_ssize_t begin, Py_ssize_t end,             # <<<<<<<<<<<<<<
 *                  int comp_ne):
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_40rfindint(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_nums, Py_ssize_t __pyx_v_begin, Py_ssize_t __pyx_v_end, int __pyx_v_comp_ne) {
  unsigned long *__pyx_v_longs;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_v_b;
  Py_ssize_t __pyx_v_e;
  Py_ssize_t __pyx_v_p;
  long __pyx_v_i;
  long __pyx_v_q;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  unsigned long __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rfindint", 0);

  /* "_gappedbuf.pyx":463
 *         cdef Py_ssize_t size, b, e, p
 * 
 *         if not (0 <= begin <= end <= self.numelems):             # <<<<<<<<<<<<<<
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))
 */
  __pyx_t_1 = (0 <= __pyx_v_begin);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_begin <= __pyx_v_end);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_end <= __pyx_v_self->numelems);
    }
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":465
 *         if not (0 <= begin <= end <= self.numelems):
 *             raise ValueError('Invalid range: %d, %d, %d' % (
 *                                 begin, end, self.numelems))             # <<<<<<<<<<<<<<
 * 
 *         size = len(nums)
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_begin); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 465; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_end); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 465; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 465; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 465; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_s_2), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_t_5));
    __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_6), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "_gappedbuf.pyx":467
 *                                 begin, end, self.numelems))
 * 
 *         size = len(nums)             # <<<<<<<<<<<<<<
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_nums); if (unlikely(__pyx_t_7 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 467; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_size = __pyx_t_7;

  /* "_gappedbuf.pyx":469
 *         size = len(nums)
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))             # <<<<<<<<<<<<<<
 *         if not longs:
 *             raise MemoryError()
 */
  __pyx_v_longs = ((unsigned long *)PyMem_Malloc((__pyx_v_size * (sizeof(unsigned long)))));

  /* "_gappedbuf.pyx":470
 *         longs = <unsigned long *>mem.PyMem_Malloc(
 *                                     size * sizeof(unsigned long))
 *         if not longs:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         try:
 */
  __pyx_t_2 = ((!(__pyx_v_longs != 0)) != 0);
  if (__pyx_t_2) {

    /* "_gappedbuf.pyx":471
 *                                     size * sizeof(unsigned long))
 *         if not longs:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         try:
 *             for i from 0 <= i < size:
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 471; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "_gappedbuf.pyx":472
 *         if not longs:
 *             raise MemoryError()
 *         try:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < size:
 *                 longs[i] = nums[i]
 */
  /*try:*/ {

    /* "_gappedbuf.pyx":473
 *             raise MemoryError()
 *         try:
 *             for i from 0 <= i < size:             # <<<<<<<<<<<<<<
 *                 longs[i] = nums[i]
 * 
 */
    __pyx_t_7 = __pyx_v_size;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_7; __pyx_v_i++) {

      /* "_gappedbuf.pyx":474
 *         try:
 *             for i from 0 <= i < size:
 *                 longs[i] = nums[i]             # <<<<<<<<<<<<<<
 * 
 *             # search after gap
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_nums, __pyx_v_i, sizeof(long), PyInt_FromLong, 0, 1, 1); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 474; __pyx_clineno = __LINE__; goto __pyx_L6;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyInt_AsUnsignedLong(__pyx_t_5); if (unlikely((__pyx_t_8 == (unsigned long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 474; __pyx_clineno = __LINE__; goto __pyx_L6;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      (__pyx_v_longs[__pyx_v_i]) = __pyx_t_8;
    }

    /* "_gappedbuf.pyx":477
 * 
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize             # <<<<<<<<<<<<<<
 *             for p from (end+self.gapsize) > p >= b:
 *                 if not comp_ne:
 */
    __pyx_t_7 = __pyx_v_self->gap;
    __pyx_t_9 = __pyx_v_begin;
    if (((__pyx_t_7 > __pyx_t_9) != 0)) {
      __pyx_t_10 = __pyx_t_7;
    } else {
      __pyx_t_10 = __pyx_t_9;
    }
    __pyx_v_b = (__pyx_t_10 + __pyx_v_self->gapsize);

    /* "_gappedbuf.pyx":478
 *             # search after gap
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from (end+self.gapsize) > p >= b:             # <<<<<<<<<<<<<<
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 */
    __pyx_t_10 = __pyx_v_b;
    for (__pyx_v_p = (__pyx_v_end + __pyx_v_self->gapsize)-1; __pyx_v_p >= __pyx_t_10; __pyx_v_p--) {

      /* "_gappedbuf.pyx":479
 *             b = max(begin, self.gap) + self.gapsize
 *             for p from (end+self.gapsize) > p >= b:
 *                 if not comp_ne:             # <<<<<<<<<<<<<<
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 */
      __pyx_t_2 = ((!(__pyx_v_comp_ne != 0)) != 0);
      if (__pyx_t_2) {

        /* "_gappedbuf.pyx":480
 *             for p from (end+self.gapsize) > p >= b:
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             return p-self.gapsize
 */
        __pyx_t_7 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_7; __pyx_v_q++) {

          /* "_gappedbuf.pyx":481
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             return p-self.gapsize
 *                 else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":482
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             return p-self.gapsize             # <<<<<<<<<<<<<<
 *                 else:
 *                     for q from 0 <= q < size:
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L6;}
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_r = __pyx_t_5;
            __pyx_t_5 = 0;
            goto __pyx_L5;
            goto __pyx_L15;
          }
          __pyx_L15:;
        }
        goto __pyx_L12;
      }
      /*else*/ {

        /* "_gappedbuf.pyx":484
 *                             return p-self.gapsize
 *                 else:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             break
 */
        __pyx_t_7 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_7; __pyx_v_q++) {

          /* "_gappedbuf.pyx":485
 *                 else:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             break
 *                     else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":486
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             break             # <<<<<<<<<<<<<<
 *                     else:
 *                         return p-self.gapsize
 */
            goto __pyx_L17_break;
            goto __pyx_L18;
          }
          __pyx_L18:;
        }
        /*else*/ {

          /* "_gappedbuf.pyx":488
 *                             break
 *                     else:
 *                         return p-self.gapsize             # <<<<<<<<<<<<<<
 * 
 *             # search before gap
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_p - __pyx_v_self->gapsize)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 488; __pyx_clineno = __LINE__; goto __pyx_L6;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L5;
        }
        __pyx_L17_break:;
      }
      __pyx_L12:;
    }

    /* "_gappedbuf.pyx":491
 * 
 *             # search before gap
 *             e = min(end, self.gap)             # <<<<<<<<<<<<<<
 *             for p from e > p >= begin:
 *                 if not comp_ne:
 */
    __pyx_t_10 = __pyx_v_self->gap;
    __pyx_t_7 = __pyx_v_end;
    if (((__pyx_t_10 < __pyx_t_7) != 0)) {
      __pyx_t_9 = __pyx_t_10;
    } else {
      __pyx_t_9 = __pyx_t_7;
    }
    __pyx_v_e = __pyx_t_9;

    /* "_gappedbuf.pyx":492
 *             # search before gap
 *             e = min(end, self.gap)
 *             for p from e > p >= begin:             # <<<<<<<<<<<<<<
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 */
    __pyx_t_9 = __pyx_v_begin;
    for (__pyx_v_p = __pyx_v_e-1; __pyx_v_p >= __pyx_t_9; __pyx_v_p--) {

      /* "_gappedbuf.pyx":493
 *             e = min(end, self.gap)
 *             for p from e > p >= begin:
 *                 if not comp_ne:             # <<<<<<<<<<<<<<
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 */
      __pyx_t_2 = ((!(__pyx_v_comp_ne != 0)) != 0);
      if (__pyx_t_2) {

        /* "_gappedbuf.pyx":494
 *             for p from e > p >= begin:
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             return p
 */
        __pyx_t_10 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

          /* "_gappedbuf.pyx":495
 *                 if not comp_ne:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             return p
 *                 else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":496
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             return p             # <<<<<<<<<<<<<<
 *                 else:
 *                     for q from 0 <= q < size:
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 496; __pyx_clineno = __LINE__; goto __pyx_L6;}
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_r = __pyx_t_5;
            __pyx_t_5 = 0;
            goto __pyx_L5;
            goto __pyx_L24;
          }
          __pyx_L24:;
        }
        goto __pyx_L21;
      }
      /*else*/ {

        /* "_gappedbuf.pyx":498
 *                             return p
 *                 else:
 *                     for q from 0 <= q < size:             # <<<<<<<<<<<<<<
 *                         if self.buf[p] == longs[q]:
 *                             break
 */
        __pyx_t_10 = __pyx_v_size;
        for (__pyx_v_q = 0; __pyx_v_q < __pyx_t_10; __pyx_v_q++) {

          /* "_gappedbuf.pyx":499
 *                 else:
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:             # <<<<<<<<<<<<<<
 *                             break
 *                     else:
 */
          __pyx_t_2 = (((__pyx_v_self->buf[__pyx_v_p]) == (__pyx_v_longs[__pyx_v_q])) != 0);
          if (__pyx_t_2) {

            /* "_gappedbuf.pyx":500
 *                     for q from 0 <= q < size:
 *                         if self.buf[p] == longs[q]:
 *                             break             # <<<<<<<<<<<<<<
 *                     else:
 *                         return p
 */
            goto __pyx_L26_break;
            goto __pyx_L27;
          }
          __pyx_L27:;
        }
        /*else*/ {

          /* "_gappedbuf.pyx":502
 *                             break
 *                     else:
 *                         return p             # <<<<<<<<<<<<<<
 *         finally:
 *             mem.PyMem_Free(longs)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_p); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 502; __pyx_clineno = __LINE__; goto __pyx_L6;}
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L5;
        }
        __pyx_L26_break:;
      }
      __pyx_L21:;
    }
  }

  /* "_gappedbuf.pyx":504
 *                         return p
 *         finally:
 *             mem.PyMem_Free(longs)             # <<<<<<<<<<<<<<
 * 
 *         return -1
 */
  /*finally:*/ {
    int __pyx_why;
    PyObject *__pyx_exc_type, *__pyx_exc_value, *__pyx_exc_tb;
    int __pyx_exc_lineno;
    __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 0; goto __pyx_L7;
    __pyx_L5: __pyx_exc_type = 0; __pyx_exc_value = 0; __pyx_exc_tb = 0; __pyx_exc_lineno = 0;
    __pyx_why = 3; goto __pyx_L7;
    __pyx_L6: {
      __pyx_why = 4;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_ErrFetch(&__pyx_exc_type, &__pyx_exc_value, &__pyx_exc_tb);
      __pyx_exc_lineno = __pyx_lineno;
      goto __pyx_L7;
    }
    __pyx_L7:;
    PyMem_Free(__pyx_v_longs);
    switch (__pyx_why) {
      case 3: goto __pyx_L0;
      case 4: {
        __Pyx_ErrRestore(__pyx_exc_type, __pyx_exc_value, __pyx_exc_tb);
        __pyx_lineno = __pyx_exc_lineno;
        __pyx_exc_type = 0;
        __pyx_exc_value = 0;
        __pyx_exc_tb = 0;
        goto __pyx_L1_error;
      }
    }
  }

  /* "_gappedbuf.pyx":506
 *             mem.PyMem_Free(longs)
 * 
 *         return -1             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_r = __pyx_int_neg_1;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.rfindint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize___get__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pxd":3
 * 
 * cdef public class GappedBuffer [object GappedBufferObj, type GappedBufferType]:
 *     cdef public Py_ssize_t bufsize             # <<<<<<<<<<<<<<
 *     cdef public Py_ssize_t numelems
 *     cdef public Py_ssize_t gap
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize___get__(struct GappedBufferObj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->bufsize); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 3; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.bufsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize_2__set__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10_gappedbuf_12GappedBuffer_7bufsize_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 3; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->bufsize = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.bufsize.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_8numelems___get__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pxd":4
 * cdef public class GappedBuffer [object GappedBufferObj, type GappedBufferType]:
 *     cdef public Py_ssize_t bufsize
 *     cdef public Py_ssize_t numelems             # <<<<<<<<<<<<<<
 *     cdef public Py_ssize_t gap
 *     cdef public Py_ssize_t gapsize
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_8numelems___get__(struct GappedBufferObj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->numelems); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 4; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.numelems.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_8numelems_2__set__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10_gappedbuf_12GappedBuffer_8numelems_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 4; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->numelems = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.numelems.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_3gap_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_3gap_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_3gap___get__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pxd":5
 *     cdef public Py_ssize_t bufsize
 *     cdef public Py_ssize_t numelems
 *     cdef public Py_ssize_t gap             # <<<<<<<<<<<<<<
 *     cdef public Py_ssize_t gapsize
 *     cdef Py_UCS4 *buf
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_3gap___get__(struct GappedBufferObj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->gap); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 5; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.gap.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_3gap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_3gap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_3gap_2__set__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10_gappedbuf_12GappedBuffer_3gap_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 5; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->gap = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.gap.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize___get__(((struct GappedBufferObj *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_gappedbuf.pxd":6
 *     cdef public Py_ssize_t numelems
 *     cdef public Py_ssize_t gap
 *     cdef public Py_ssize_t gapsize             # <<<<<<<<<<<<<<
 *     cdef Py_UCS4 *buf
 * 
 */

static PyObject *__pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize___get__(struct GappedBufferObj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->gapsize); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 6; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.gapsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize_2__set__(((struct GappedBufferObj *)__pyx_v_self), ((PyObject *)__pyx_v_value));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10_gappedbuf_12GappedBuffer_7gapsize_2__set__(struct GappedBufferObj *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 6; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->gapsize = __pyx_t_1;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("_gappedbuf.GappedBuffer.gapsize.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10_gappedbuf_GappedBuffer __pyx_vtable_10_gappedbuf_GappedBuffer;

static PyObject *__pyx_tp_new_10_gappedbuf_GappedBuffer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct GappedBufferObj *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct GappedBufferObj *)o);
  p->__pyx_vtab = __pyx_vtabptr_10_gappedbuf_GappedBuffer;
  if (unlikely(__pyx_pw_10_gappedbuf_12GappedBuffer_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_10_gappedbuf_GappedBuffer(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_10_gappedbuf_12GappedBuffer_3__dealloc__(o);
    if (PyErr_Occurred()) PyErr_WriteUnraisable(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_10_gappedbuf_GappedBuffer(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_10_gappedbuf_GappedBuffer(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_19__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_21__delitem__(o, i);
  }
}

static PyObject *__pyx_getprop_10_gappedbuf_12GappedBuffer_bufsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_1__get__(o);
}

static int __pyx_setprop_10_gappedbuf_12GappedBuffer_bufsize(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_7bufsize_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10_gappedbuf_12GappedBuffer_numelems(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_1__get__(o);
}

static int __pyx_setprop_10_gappedbuf_12GappedBuffer_numelems(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_8numelems_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10_gappedbuf_12GappedBuffer_gap(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10_gappedbuf_12GappedBuffer_3gap_1__get__(o);
}

static int __pyx_setprop_10_gappedbuf_12GappedBuffer_gap(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_3gap_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10_gappedbuf_12GappedBuffer_gapsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_1__get__(o);
}

static int __pyx_setprop_10_gappedbuf_12GappedBuffer_gapsize(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10_gappedbuf_12GappedBuffer_7gapsize_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_10_gappedbuf_GappedBuffer[] = {
  {__Pyx_NAMESTR("insert"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_5insert, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("append"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_7append, METH_O, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("delete"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_9delete, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("replace"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_11replace, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("get"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_13get, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("findchr"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_23findchr, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("rfindchr"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_25rfindchr, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("getint"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_27getint, METH_O, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("getints"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_29getints, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("insertints"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_31insertints, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("appendints"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_33appendints, METH_O, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("replaceints"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_35replaceints, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("setints"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_37setints, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("findint"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_39findint, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("rfindint"), (PyCFunction)__pyx_pw_10_gappedbuf_12GappedBuffer_41rfindint, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10_gappedbuf_GappedBuffer[] = {
  {(char *)"bufsize", __pyx_getprop_10_gappedbuf_12GappedBuffer_bufsize, __pyx_setprop_10_gappedbuf_12GappedBuffer_bufsize, 0, 0},
  {(char *)"numelems", __pyx_getprop_10_gappedbuf_12GappedBuffer_numelems, __pyx_setprop_10_gappedbuf_12GappedBuffer_numelems, 0, 0},
  {(char *)"gap", __pyx_getprop_10_gappedbuf_12GappedBuffer_gap, __pyx_setprop_10_gappedbuf_12GappedBuffer_gap, 0, 0},
  {(char *)"gapsize", __pyx_getprop_10_gappedbuf_12GappedBuffer_gapsize, __pyx_setprop_10_gappedbuf_12GappedBuffer_gapsize, 0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_GappedBuffer = {
  __pyx_pw_10_gappedbuf_12GappedBuffer_15__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10_gappedbuf_GappedBuffer, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GappedBuffer = {
  __pyx_pw_10_gappedbuf_12GappedBuffer_15__len__, /*mp_length*/
  __pyx_pw_10_gappedbuf_12GappedBuffer_17__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_10_gappedbuf_GappedBuffer, /*mp_ass_subscript*/
};

DL_EXPORT(PyTypeObject) GappedBufferType = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("_gappedbuf.GappedBuffer"), /*tp_name*/
  sizeof(struct GappedBufferObj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10_gappedbuf_GappedBuffer, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_GappedBuffer, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GappedBuffer, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10_gappedbuf_GappedBuffer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10_gappedbuf_GappedBuffer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10_gappedbuf_GappedBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
  #if PY_VERSION_HEX >= 0x030400a1 && defined(Py_TPFLAGS_HAVE_FINALIZE)
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    __Pyx_NAMESTR("_gappedbuf"),
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 1, 0},
  {&__pyx_kp_s_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 0, 1, 0},
  {&__pyx_kp_s_3, __pyx_k_3, sizeof(__pyx_k_3), 0, 0, 1, 0},
  {&__pyx_kp_s_5, __pyx_k_5, sizeof(__pyx_k_5), 0, 0, 1, 0},
  {&__pyx_kp_s_8, __pyx_k_8, sizeof(__pyx_k_8), 0, 0, 1, 0},
  {&__pyx_n_s__IndexError, __pyx_k__IndexError, sizeof(__pyx_k__IndexError), 0, 0, 1, 1},
  {&__pyx_n_s__MemoryError, __pyx_k__MemoryError, sizeof(__pyx_k__MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s__RuntimeError, __pyx_k__RuntimeError, sizeof(__pyx_k__RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s__ValueError, __pyx_k__ValueError, sizeof(__pyx_k__ValueError), 0, 0, 1, 1},
  {&__pyx_n_s____import__, __pyx_k____import__, sizeof(__pyx_k____import__), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____pyx_vtable__, __pyx_k____pyx_vtable__, sizeof(__pyx_k____pyx_vtable__), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s___gappedbuf, __pyx_k___gappedbuf, sizeof(__pyx_k___gappedbuf), 0, 0, 1, 1},
  {&__pyx_n_s___gappedbufre, __pyx_k___gappedbufre, sizeof(__pyx_k___gappedbufre), 0, 0, 1, 1},
  {&__pyx_n_s__append, __pyx_k__append, sizeof(__pyx_k__append), 0, 0, 1, 1},
  {&__pyx_n_s__appendints, __pyx_k__appendints, sizeof(__pyx_k__appendints), 0, 0, 1, 1},
  {&__pyx_n_s__begin, __pyx_k__begin, sizeof(__pyx_k__begin), 0, 0, 1, 1},
  {&__pyx_n_s__c, __pyx_k__c, sizeof(__pyx_k__c), 0, 0, 1, 1},
  {&__pyx_n_s__comp_ne, __pyx_k__comp_ne, sizeof(__pyx_k__comp_ne), 0, 0, 1, 1},
  {&__pyx_n_s__delete, __pyx_k__delete, sizeof(__pyx_k__delete), 0, 0, 1, 1},
  {&__pyx_n_s__end, __pyx_k__end, sizeof(__pyx_k__end), 0, 0, 1, 1},
  {&__pyx_n_s__get, __pyx_k__get, sizeof(__pyx_k__get), 0, 0, 1, 1},
  {&__pyx_n_s__getint, __pyx_k__getint, sizeof(__pyx_k__getint), 0, 0, 1, 1},
  {&__pyx_n_s__getints, __pyx_k__getints, sizeof(__pyx_k__getints), 0, 0, 1, 1},
  {&__pyx_n_s__index, __pyx_k__index, sizeof(__pyx_k__index), 0, 0, 1, 1},
  {&__pyx_n_s__init_gappedbufre, __pyx_k__init_gappedbufre, sizeof(__pyx_k__init_gappedbufre), 0, 0, 1, 1},
  {&__pyx_n_s__insert, __pyx_k__insert, sizeof(__pyx_k__insert), 0, 0, 1, 1},
  {&__pyx_n_s__insertints, __pyx_k__insertints, sizeof(__pyx_k__insertints), 0, 0, 1, 1},
  {&__pyx_n_s__modules, __pyx_k__modules, sizeof(__pyx_k__modules), 0, 0, 1, 1},
  {&__pyx_n_s__nums, __pyx_k__nums, sizeof(__pyx_k__nums), 0, 0, 1, 1},
  {&__pyx_n_s__replace, __pyx_k__replace, sizeof(__pyx_k__replace), 0, 0, 1, 1},
  {&__pyx_n_s__replaceints, __pyx_k__replaceints, sizeof(__pyx_k__replaceints), 0, 0, 1, 1},
  {&__pyx_n_s__s, __pyx_k__s, sizeof(__pyx_k__s), 0, 0, 1, 1},
  {&__pyx_n_s__setints, __pyx_k__setints, sizeof(__pyx_k__setints), 0, 0, 1, 1},
  {&__pyx_n_s__sys, __pyx_k__sys, sizeof(__pyx_k__sys), 0, 0, 1, 1},
  {&__pyx_n_s__v, __pyx_k__v, sizeof(__pyx_k__v), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s__MemoryError); if (!__pyx_builtin_MemoryError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 24; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s__ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 84; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s__IndexError); if (!__pyx_builtin_IndexError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s__RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 207; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "_gappedbuf.pyx":222
 *                      &step, &slicelength)
 *             if step != 1:
 *                 raise ValueError('step is not supported')             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_k_tuple_4 = PyTuple_Pack(1, ((PyObject *)__pyx_kp_s_3)); if (unlikely(!__pyx_k_tuple_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_4);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_4));

  /* "_gappedbuf.pyx":14
 * 
 * 
 * def init_gappedbufre():             # <<<<<<<<<<<<<<
 *     import sys
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()
 */
  __pyx_k_tuple_6 = PyTuple_Pack(1, ((PyObject *)__pyx_n_s__sys)); if (unlikely(!__pyx_k_tuple_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 14; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_6);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_6));
  __pyx_k_codeobj_7 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_8, __pyx_n_s__init_gappedbufre, 14, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 14; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_gappedbuf(void); /*proto*/
PyMODINIT_FUNC init_gappedbuf(void)
#else
PyMODINIT_FUNC PyInit__gappedbuf(void); /*proto*/
PyMODINIT_FUNC PyInit__gappedbuf(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__gappedbuf(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("_gappedbuf"), __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "_gappedbuf")) {
      if (unlikely(PyDict_SetItemString(modules, "_gappedbuf", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main__gappedbuf) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_10_gappedbuf_GappedBuffer = &__pyx_vtable_10_gappedbuf_GappedBuffer;
  __pyx_vtable_10_gappedbuf_GappedBuffer._resize_buf = (void (*)(struct GappedBufferObj *, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__resize_buf;
  __pyx_vtable_10_gappedbuf_GappedBuffer._expand_gap = (void (*)(struct GappedBufferObj *, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__expand_gap;
  __pyx_vtable_10_gappedbuf_GappedBuffer._shrink_gap = (void (*)(struct GappedBufferObj *))__pyx_f_10_gappedbuf_12GappedBuffer__shrink_gap;
  __pyx_vtable_10_gappedbuf_GappedBuffer._move_gap = (void (*)(struct GappedBufferObj *, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__move_gap;
  __pyx_vtable_10_gappedbuf_GappedBuffer._get_max = (Py_UCS4 (*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__get_max;
  __pyx_vtable_10_gappedbuf_GappedBuffer._copy_chars1 = (void (*)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__copy_chars1;
  __pyx_vtable_10_gappedbuf_GappedBuffer._copy_chars2 = (void (*)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__copy_chars2;
  __pyx_vtable_10_gappedbuf_GappedBuffer._copy_chars4 = (void (*)(struct GappedBufferObj *, void *, Py_ssize_t, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__copy_chars4;
  __pyx_vtable_10_gappedbuf_GappedBuffer._get_slice = (int (*)(struct GappedBufferObj *, PyObject *, Py_ssize_t *, Py_ssize_t *))__pyx_f_10_gappedbuf_12GappedBuffer__get_slice;
  __pyx_vtable_10_gappedbuf_GappedBuffer._insert = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, PyObject *))__pyx_f_10_gappedbuf_12GappedBuffer__insert;
  __pyx_vtable_10_gappedbuf_GappedBuffer.insert = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_insert;
  __pyx_vtable_10_gappedbuf_GappedBuffer.append = (PyObject *(*)(struct GappedBufferObj *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_append;
  __pyx_vtable_10_gappedbuf_GappedBuffer._delete = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t))__pyx_f_10_gappedbuf_12GappedBuffer__delete;
  __pyx_vtable_10_gappedbuf_GappedBuffer.delete = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_delete;
  __pyx_vtable_10_gappedbuf_GappedBuffer.replace = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_replace;
  __pyx_vtable_10_gappedbuf_GappedBuffer.get = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_get;
  __pyx_vtable_10_gappedbuf_GappedBuffer.getint = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_getint;
  __pyx_vtable_10_gappedbuf_GappedBuffer.getints = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_getints;
  __pyx_vtable_10_gappedbuf_GappedBuffer._insertints = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, PyObject *))__pyx_f_10_gappedbuf_12GappedBuffer__insertints;
  __pyx_vtable_10_gappedbuf_GappedBuffer.insertints = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_insertints;
  __pyx_vtable_10_gappedbuf_GappedBuffer.appendints = (PyObject *(*)(struct GappedBufferObj *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_appendints;
  __pyx_vtable_10_gappedbuf_GappedBuffer.replaceints = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_replaceints;
  __pyx_vtable_10_gappedbuf_GappedBuffer.setints = (PyObject *(*)(struct GappedBufferObj *, Py_ssize_t, Py_ssize_t, unsigned long, int __pyx_skip_dispatch))__pyx_f_10_gappedbuf_12GappedBuffer_setints;
  if (PyType_Ready(&GappedBufferType) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(GappedBufferType.tp_dict, __pyx_vtabptr_10_gappedbuf_GappedBuffer) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "GappedBuffer", (PyObject *)&GappedBufferType) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_10_gappedbuf_GappedBuffer = &GappedBufferType;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) {__pyx_filename = __pyx_f[3]; __pyx_lineno = 8; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) {__pyx_filename = __pyx_f[4]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "_gappedbuf.pyx":14
 * 
 * 
 * def init_gappedbufre():             # <<<<<<<<<<<<<<
 *     import sys
 *     sys.modules['_gappedbufre'] = PyInit__gappedbufre()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_10_gappedbuf_1init_gappedbufre, NULL, __pyx_n_s___gappedbuf); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 14; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s__init_gappedbufre, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 14; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_gappedbuf.pyx":1
 * from libc.stdlib cimport malloc, free, realloc             # <<<<<<<<<<<<<<
 * from libc.string cimport memmove, memcpy
 * from cpython cimport unicode, mem, ref
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  if (PyDict_SetItem(__pyx_d, __pyx_n_s____test__, ((PyObject *)__pyx_t_1)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    __Pyx_AddTraceback("init _gappedbuf", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init _gappedbuf");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CPYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
        if (PyInstance_Check(type)) {
            type = (PyObject*) ((PyInstanceObject*)type)->in_class;
            Py_INCREF(type);
        } else {
            type = 0;
            PyErr_SetString(PyExc_TypeError,
                "raise: exception must be an old-style class or instance");
            goto raise_error;
        }
        #else
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        } else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (!type) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (Py_TYPE(obj) == type) return 1;
    }
    else {
        if (PyObject_TypeCheck(obj, type)) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%s' has incorrect type (expected %s, got %s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (PyErr_ExceptionMatches(PyExc_OverflowError))
                        PyErr_Clear();
                    else
                        return NULL;
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION==3&&PY_MINOR_VERSION==0)
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s____pyx_vtable__, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s____import__);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

static CYTHON_INLINE Py_UCS4 __Pyx_PyObject_AsPy_UCS4(PyObject* x) {
   long ival;
   if (PyUnicode_Check(x)) {
       Py_ssize_t length;
       #if CYTHON_PEP393_ENABLED
       length = PyUnicode_GET_LENGTH(x);
       if (likely(length == 1)) {
           return PyUnicode_READ_CHAR(x, 0);
       }
       #else
       length = PyUnicode_GET_SIZE(x);
       if (likely(length == 1)) {
           return PyUnicode_AS_UNICODE(x)[0];
       }
       #if Py_UNICODE_SIZE == 2
       else if (PyUnicode_GET_SIZE(x) == 2) {
           Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
           if (high_val >= 0xD800 && high_val <= 0xDBFF) {
               Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
               if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
                   return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
               }
           }
       }
       #endif
       #endif
       PyErr_Format(PyExc_ValueError,
                    "only single character unicode strings can be converted to Py_UCS4, "
                    "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
       return (Py_UCS4)-1;
   }
   ival = __Pyx_PyInt_AsLong(x);
   if (unlikely(ival < 0)) {
       if (!PyErr_Occurred())
           PyErr_SetString(PyExc_OverflowError,
                           "cannot convert negative value to Py_UCS4");
       return (Py_UCS4)-1;
   } else if (unlikely(ival > 1114111)) {
       PyErr_SetString(PyExc_OverflowError,
                       "value too large to convert to Py_UCS4");
       return (Py_UCS4)-1;
   }
   return (Py_UCS4)ival;
}

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (unsigned long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(unsigned long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(unsigned long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(unsigned PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(unsigned PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (signed long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed long)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(signed long) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(signed long) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
#include "longintrepr.h"
#endif
#endif
static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return (signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
#if CYTHON_USE_PYLONG_INTERNALS
            if (sizeof(digit) <= sizeof(signed PY_LONG_LONG)) {
                switch (Py_SIZE(x)) {
                    case  0: return 0;
                    case  1: return +(signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                    case -1: return -(signed PY_LONG_LONG) ((PyLongObject*)x)->ob_digit[0];
                }
            }
#endif
#endif
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif /*__PYX_DEFAULT_STRING_ENCODING_IS_ASCII*/
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else /* PY_VERSION_HEX < 0x03030000 */
        if (PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_DATA_SIZE(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII */
        return PyUnicode_AsUTF8AndSize(o, length);
#endif /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII */
#endif /* PY_VERSION_HEX < 0x03030000 */
    } else
#endif /* __PYX_DEFAULT_STRING_ENCODING_IS_ASCII  || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT */
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (r < 0) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       if ((val != (unsigned PY_LONG_LONG)-1) || !PyErr_Occurred())
           PyErr_SetString(PyExc_OverflowError,
                           "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
